/* $Id: globus-replica-catalog.c,v 1.23 2003/03/31 23:04:28 link Exp $ */

/**
 * @file
 *
 * Globus Replica Catalog Tool Source File
 *
 * This file defines a command line tool to manage the information in a
 * replica catalog.
 *
 * @see globus_replica_catalog.h
 */

#include "globus_common.h"
#include "globus_args.h"
#include "globus_replica_catalog.h"
#include "version.h" /* provides local_version */

#define GLOBUS_REPLICA_CATALOG_BUFSIZE 256

const char * usage_short =
"globus-replica-catalog  HOST OBJECT ACTION\n";

const char* usage_long =
"\n"
"globus-replica-catalog HOST OBJECT ACTION\n"
"globus-replica-catalog -version[s]\n"
"globus-replica-catalog -help\n"
"\n"
"  HOST\n"
"    -host <collection url>\n"
"       connect to a host of the given LDAP URL of the form\n"
"       ldap://host[:port]/dn which includes the DN of a collection, can\n"
"       be set with the GLOBUS_REPLICA_CATALOG_HOST environment variable\n"
"    [ -manager <manager DN> ]\n"
"       DN to be used during authentication, can be set with the\n"
"       GLOBUS_REPLICA_CATALOG_MANAGER environment variable\n"
"    [ -password <input file> ] \n"
"       password to be used during authentication\n"
"\n"
"  OBJECT\n"
"    -collection\n"
"       act on the collection given in the DN of the LDAP URL\n"
"    -location <location name>\n"
"       act on the given location name\n"
"    -logicalfile <logical file name>\n"
"       act on the given logicalfile name\n"
"\n"
"  ACTION\n"
"    -create [ <input file> ] | -create <location url> [ <input file> ]\n"
"          | -create <size>\n"
"       create a collection with filenames in the given input file,\n"
"       a location with the given location url and optional set of\n"
"       filenames in the given input file, or a logicalfile with the\n"
"       given size\n"
"    -delete\n"
"       delete a collection, location, or logical file\n"
"    -add-attribute <attribute name> <attribute value>\n"
"       add the given attribute name and value to a collection\n"
"       location, or logical file\n"
"    -delete-attribute <attribute name> [ <attribute value> ]\n"
"       delete an attribute or optionally the specific value of an\n"
"       attribute from a collection, location, or logica file\n"
"    -list-attributes [ <attribute names> ]\n"
"       list attributes of a collection, location or logical file,\n"
"       returning all attributes or optionally only the given attributes\n"
"    -add-filenames <input file> [ -add-anyway ]\n"
"       add filenames in the given input file to a collection or\n"
"       location, optionally add the filenames that aren't there if some\n"
"       of the filenames are already there anyway\n"
"    -delete-filenames <input file>\n"
"       delete filenames in the given input file from a collection\n"
"       or location\n"
"    -list-filenames\n"
"       list filenames in a collection or location\n"
"    -list-locations [ <attribute names> ]\n"
"       list locations in a collection, optionally restrict the\n"
"       attributes returned to the attribute names given\n"
"    -list-logicalfiles [ <attribute names> ]\n"
"       list logical files in a collection, optionally restrict the\n"
"       attributes returned to the attribute names given\n"
"    -find-locations <input file> [ -attributes <attribute names> ]\n"
"       find locations in a collection that have filenames in the given\n"
"       input file, optionally restrict the attributes returned to the\n"
"       attribute names given\n"
"\n"
"    -list-collections\n"
"       list collections in a replica catalog, the DN in the given host\n"
"       LDAP URL is the base from which to search, no object need be\n"
"       specified\n"
"\n"
"  Example\n"
"    globus-replica-catalog \\\n"
"       -host \"ldap://ldap2.mcs.anl.gov/lc=B04.10.atm,\\\n"
"               rc=PCMDI Replica Catalog,o=Argonne National Laboratory,\\\n"
"               o=Globus,c=US\"\\\n"
"       -collection -list-filenames\n"
"\n";

enum { arg_host = 1, arg_manager, arg_password,
	arg_collection, arg_location, arg_logicalfile,
	arg_create, arg_delete,
	arg_add_attribute, arg_delete_attribute, arg_list_attributes,
	arg_add_filenames, arg_delete_filenames, arg_list_filenames,
	arg_add_anyway,
	arg_list_locations, arg_list_logicalfiles, arg_find_locations,
	arg_list_collections,
	arg_attributes,
	optc = arg_attributes };

static char * arg_host_aliases[] = {
	"-host", "-h",
	GLOBUS_NULL };
static char * arg_manager_aliases[] = {
	"-manager", "-m", "-D",
	GLOBUS_NULL };
static char * arg_password_aliases[] = {
	"-password", "-P",
	GLOBUS_NULL };
static char * arg_collection_aliases[] = {
	"-collection", "-coll", "-c", 
	GLOBUS_NULL };
static char * arg_location_aliases[] = {
	"-location", "-loc", "-l",
	GLOBUS_NULL };
static char * arg_logicalfile_aliases[] = {
	"-logicalfile", "-file", "-lf", "-f",
	GLOBUS_NULL};
static char * arg_create_aliases[] = {
	"-create",
	GLOBUS_NULL };
static char * arg_delete_aliases[] = {
	"-delete",
	GLOBUS_NULL };
static char * arg_add_attribute_aliases[] = {
	"-add-attribute", "-add-attr",
	GLOBUS_NULL };
static char * arg_delete_attribute_aliases[] = {
	"-delete-attribute", "-delete-attr",
	GLOBUS_NULL };
static char * arg_list_attributes_aliases[] = {
	"-list-attributes", "-list-attrs",
	GLOBUS_NULL };
static char * arg_add_filenames_aliases[] = {
	"-add-filenames",
	GLOBUS_NULL };
static char * arg_delete_filenames_aliases[] = {
	"-delete-filenames",
	GLOBUS_NULL };
static char * arg_list_filenames_aliases[] = {
	"-list-filenames",
	GLOBUS_NULL };
static char * arg_add_anyway_aliases[] = {
	"-add-anyway", "-add",
	GLOBUS_NULL };
static char * arg_list_locations_aliases[] = {
	"-list-locations",
	GLOBUS_NULL };
static char * arg_list_logicalfiles_aliases[] = {
	"-list-logicalfiles",
	GLOBUS_NULL };
static char * arg_find_locations_aliases[] = {
	"-find-locations", "-find",
	GLOBUS_NULL };
static char * arg_list_collections_aliases[] = {
	"-list-collections",
	GLOBUS_NULL };
static char * arg_attributes_aliases[] = {
	"-attributes", "-attrs", "-attribute-names", "-attrnames",
	GLOBUS_NULL };

static globus_args_option_descriptor_t optv[] = {
	{ arg_host, arg_host_aliases, 1,
	    GLOBUS_NULL, GLOBUS_NULL },
	{ arg_manager, arg_manager_aliases, 1,
	    GLOBUS_NULL, GLOBUS_NULL },
	{ arg_password, arg_password_aliases, 1,
	    GLOBUS_NULL, GLOBUS_NULL },
	{ arg_collection, arg_collection_aliases, 0,
	    GLOBUS_NULL, GLOBUS_NULL },
	{ arg_location, arg_location_aliases, 1,
	    GLOBUS_NULL, GLOBUS_NULL },
	{ arg_logicalfile, arg_logicalfile_aliases, 1,
	    GLOBUS_NULL, GLOBUS_NULL },
	{ arg_create, arg_create_aliases, 0,
	    GLOBUS_NULL, GLOBUS_NULL },
	{ arg_delete, arg_delete_aliases, 0,
	    GLOBUS_NULL, GLOBUS_NULL },
	{ arg_add_attribute, arg_add_attribute_aliases, 2,
	    GLOBUS_NULL, GLOBUS_NULL },
	{ arg_delete_attribute, arg_delete_attribute_aliases, 1,
	    GLOBUS_NULL, GLOBUS_NULL },
	{ arg_list_attributes, arg_list_attributes_aliases, 0,
	    GLOBUS_NULL, GLOBUS_NULL },
	{ arg_add_filenames, arg_add_filenames_aliases, 1,
	    GLOBUS_NULL, GLOBUS_NULL },
	{ arg_delete_filenames, arg_delete_filenames_aliases, 1,
	    GLOBUS_NULL, GLOBUS_NULL },
	{ arg_list_filenames, arg_list_filenames_aliases, 0,
	    GLOBUS_NULL, GLOBUS_NULL },
	{ arg_add_anyway, arg_add_anyway_aliases, 0,
	    GLOBUS_NULL, GLOBUS_NULL },
	{ arg_list_locations, arg_list_locations_aliases, 0,
	    GLOBUS_NULL, GLOBUS_NULL },
	{ arg_list_logicalfiles, arg_list_logicalfiles_aliases, 0,
	    GLOBUS_NULL, GLOBUS_NULL },
	{ arg_find_locations, arg_find_locations_aliases, 1,
	    GLOBUS_NULL, GLOBUS_NULL },
	{ arg_list_collections, arg_list_collections_aliases, 0,
	    GLOBUS_NULL, GLOBUS_NULL },
	{ arg_attributes, arg_attributes_aliases, 0,
	    GLOBUS_NULL, GLOBUS_NULL } };

#define globus_l_replica_catalog_action_error(err) \
{ \
    globus_libc_fprintf( stderr, \
	    globus_object_printable_to_string( \
		    globus_error_get(err) ) ); \
    globus_module_deactivate_all(); \
    exit(1); \
}

#define globus_l_replica_catalog_arg_error(msg) \
{ \
    globus_libc_fprintf( stderr, \
	    "%s\nUse -help to display full usage.\n", (msg) ); \
    globus_module_deactivate_all(); \
    exit(1); \
}

#define globus_l_replica_catalog_object_def(object, arg) \
{ \
    if (object != 0) { \
	globus_l_replica_catalog_arg_error( \
	"Specify no more than one object to act on." ); \
    } else { \
	object = arg; \
    } \
}

#define globus_l_replica_catalog_action_def(action, arg) \
{ \
    if (action != 0) { \
	globus_l_replica_catalog_arg_error( \
	"Specify no more than one action." ); \
    } else { \
	action = arg; \
    } \
}

void
globus_l_replica_catalog_print_filenames(
	char ** filenames )
{
    int i;

    if (filenames == GLOBUS_NULL) {
	return;
    }

    for (i = 0; filenames[i]; i++ ){
	globus_libc_printf( "%s\n", filenames[i] );
    }
}

void
globus_l_replica_catalog_print_attributes(
	globus_replica_catalog_attribute_set_t attributes )
{
    char * name;
    char ** values;
    int i;
    
    globus_replica_catalog_attribute_set_first(&attributes);
    while (globus_replica_catalog_attribute_set_more(&attributes)) {
	globus_replica_catalog_attribute_set_get_name( &attributes,
		&name );
	globus_replica_catalog_attribute_set_get_values( &attributes,
		&values );
	for (i = 0; values[i]; i++ ){
	    globus_libc_printf("%s=%s\n", name, values[i]);
	}
	globus_libc_free(name);
	globus_replica_catalog_attribute_set_free_values(values);
	globus_replica_catalog_attribute_set_next(&attributes);
    }
}

void
globus_l_replica_catalog_print_entries(
	globus_replica_catalog_entry_set_t entries )
{
    globus_replica_catalog_attribute_set_t attributes;

    globus_replica_catalog_entry_set_first(&entries);
    while (globus_replica_catalog_entry_set_more(&entries)) {
	globus_replica_catalog_attribute_set_init(&attributes);
	globus_replica_catalog_entry_set_get_attributes( &entries,
		&attributes );
	globus_l_replica_catalog_print_attributes(attributes);
	globus_libc_printf("\n");
	globus_replica_catalog_attribute_set_destroy(&attributes);
	globus_replica_catalog_entry_set_next(&entries);
    }
}

void
globus_l_replica_catalog_print_entry_dns(
	globus_replica_catalog_entry_set_t entries )
{
    char * dn;

    globus_replica_catalog_entry_set_first(&entries);
    while (globus_replica_catalog_entry_set_more(&entries)) {
	globus_replica_catalog_entry_set_get_dn(&entries, &dn);
	globus_libc_printf("%s\n", dn);
	globus_replica_catalog_entry_set_free_dn(dn);
	globus_replica_catalog_entry_set_next(&entries);
    }
}

int
main(
	int argc,
	char *argv[] )
{
    globus_list_t * options_found = GLOBUS_NULL;
    globus_list_t * list = GLOBUS_NULL;
    globus_args_option_instance_t * option = GLOBUS_NULL;
    int object = 0;
    int action = 0;
    char * collection_url = GLOBUS_NULL;
    char * name = GLOBUS_NULL;
    char * filenames_file = GLOBUS_NULL;
    char * password_file = GLOBUS_NULL;
    char ** filenames = GLOBUS_NULL;
    char * size = GLOBUS_NULL;
    char * location_url = GLOBUS_NULL;
    char * attribute_name = GLOBUS_NULL;
    char * attribute_value = GLOBUS_NULL;
    char ** attribute_names = GLOBUS_NULL;
    globus_bool_t add_anyway_p = GLOBUS_FALSE;
    globus_bool_t attributes_p = GLOBUS_FALSE;
    char * manager = GLOBUS_NULL;
    char * password = GLOBUS_NULL;
    int i;
    FILE *fd;
    globus_replica_catalog_collection_handleattr_t handleattrs;
    globus_replica_catalog_collection_handle_t handle;
    globus_replica_catalog_attribute_set_t attributes;
    globus_replica_catalog_entry_set_t entries;
    globus_result_t * err;

    if ( globus_module_activate(GLOBUS_REPLICA_CATALOG_MODULE) !=
	    GLOBUS_SUCCESS ) {
	globus_libc_fprintf( stderr,
		"Error activating the replica catalog module.\n" );
	exit(1);
    }

    /*
     * scan arguments
     */

    if ( (i = globus_args_scan( &argc, &argv,
	    optc, optv,
	    "globus-replica-catalog-program", &local_version,
	    usage_short, usage_long, 
	    &options_found, GLOBUS_NULL )) < 0 ) {
	globus_module_deactivate_all();
	exit(i == GLOBUS_FAILURE ? 1 : 0);
    }

    for ( list = options_found;
	    !globus_list_empty(list);
	    list = globus_list_rest(list) ) {
	option = globus_list_first(list);
	switch (option->id_number) {
	    case arg_host:
		collection_url = globus_libc_strdup(option->values[0]);
		break;
	    case arg_manager:
		manager = globus_libc_strdup(option->values[0]);
		break;
	    case arg_password:
		password_file = globus_libc_strdup(option->values[0]);
		break;
	    case arg_collection:
		globus_l_replica_catalog_object_def( object,
			arg_collection );
		break;
	    case arg_location:
		globus_l_replica_catalog_object_def(object, arg_location);
		name = globus_libc_strdup(option->values[0]);
		break;
	    case arg_logicalfile:
		globus_l_replica_catalog_object_def( object,
			arg_logicalfile );
		name = globus_libc_strdup(option->values[0]);
		break;
	    case arg_create:
		globus_l_replica_catalog_action_def(action, arg_create);
		break;
	    case arg_delete:
		globus_l_replica_catalog_action_def(action, arg_delete);
		break;
	    case arg_add_attribute:
		globus_l_replica_catalog_action_def( action,
			arg_add_attribute );
		attribute_name = globus_libc_strdup(option->values[0]);
		attribute_value = globus_libc_strdup(option->values[1]);
		break;
	    case arg_delete_attribute:
		globus_l_replica_catalog_action_def( action,
			arg_delete_attribute );
		attribute_name = globus_libc_strdup(option->values[0]);
		if (argc > 1) {
		    attribute_value = globus_libc_strdup(argv[1]);
		}
		break;
	    case arg_list_attributes:
		globus_l_replica_catalog_action_def( action,
			arg_list_attributes );
		break;
	    case arg_add_filenames:
		globus_l_replica_catalog_action_def( action,
			arg_add_filenames );
		filenames_file = globus_libc_strdup(option->values[0]);
		break;
	    case arg_delete_filenames:
		globus_l_replica_catalog_action_def( action,
			arg_delete_filenames );
		filenames_file = globus_libc_strdup(option->values[0]);
		break;
	    case arg_list_filenames:
		globus_l_replica_catalog_action_def( action,
			arg_list_filenames );
		break;
	    case arg_add_anyway:
		add_anyway_p = GLOBUS_TRUE;
		break;
	    case arg_list_locations:
		globus_l_replica_catalog_action_def( action,
			arg_list_locations );
		break;
	    case arg_list_logicalfiles:
		globus_l_replica_catalog_action_def( action,
			arg_list_logicalfiles );
		break;
	    case arg_find_locations:
		globus_l_replica_catalog_action_def( action,
			arg_find_locations );
		filenames_file = globus_libc_strdup(option->values[0]);
		break;
	    case arg_list_collections:
		globus_l_replica_catalog_action_def( action,
			arg_list_collections );
		break;
	    case arg_attributes:
		attributes_p = GLOBUS_TRUE;
		break;
	    default:
		globus_l_replica_catalog_arg_error(
		"Error scanning arguments." );
		break;
	}
    }

    globus_args_option_instance_list_free(&options_found);

    /*
     * check arguments
     */

    if (collection_url == GLOBUS_NULL) {
	if ( ( collection_url = globus_libc_getenv(
		"GLOBUS_REPLICA_CATALOG_HOST" ) ) == GLOBUS_NULL ) {
	    globus_l_replica_catalog_arg_error(
	    "Specify -host." );
	}
    }

    if ((object == 0) && (action != arg_list_collections)) {
	globus_l_replica_catalog_arg_error(
	"Specify an object." );
    }

    if (action == 0) {
	globus_l_replica_catalog_arg_error(
	"Specify an action." );
    }

    if ((action == arg_list_locations) && (object != arg_collection)) {
	globus_l_replica_catalog_arg_error(
	"Specify -list-locations with -collection only." );
    }

    if ((action == arg_list_logicalfiles) && (object != arg_collection)) {
	globus_l_replica_catalog_arg_error(
	"Specify -list-logicalfiles with -collection only." );
    }

    if ((action == arg_find_locations) && (object != arg_collection)) {
	globus_l_replica_catalog_arg_error(
	"Specify -find-locations with -collection only." );
    }

    if ( ( (action == arg_add_filenames)
	    || (action == arg_delete_filenames)
	    || (action == arg_list_filenames) )
	    && (object == arg_logicalfile) ) {
	globus_l_replica_catalog_arg_error(
	"Specify -*-filenames with -collection and -location only." );
    }

    if ((add_anyway_p == GLOBUS_TRUE) && (action != arg_add_filenames)) {
	globus_l_replica_catalog_arg_error(
	"Specify -add-anyway with -add-filenames only." );
    }
 
    if ((attributes_p == GLOBUS_TRUE) && (action != arg_find_locations)) {
	globus_l_replica_catalog_arg_error(
	"Specify -attributes with -find_locations only." );
    }

    if ((attributes_p == GLOBUS_TRUE) && !(argc > 1)) {
	globus_l_replica_catalog_arg_error(
	"Specify -attributes with at least one attribute name." );
    }

    if ( ( (action == arg_add_attribute)
	    || (action == arg_delete_attribute)
	    || (action == arg_add_filenames)
	    || (action == arg_delete_filenames)
	    || (action == arg_create)
	    || (action == arg_delete) ) 
	    && (manager == GLOBUS_NULL) ) {
	if ( ( manager = globus_libc_getenv(
		"GLOBUS_REPLICA_CATALOG_MANAGER" ) ) == GLOBUS_NULL ) {
	    globus_l_replica_catalog_arg_error(
	    "Specify -manager with this action." );
	}
    }
    
    /*
     * get extra arguments
     */

    if (manager != GLOBUS_NULL) {
	if (password_file == GLOBUS_NULL) {

	    password = getpass("Password: ");

	} else {

	    if (strcmp(password_file,"-") == 0) {
		fd = stdin;
	    } else if ((fd = fopen(password_file, "r")) == NULL) {
		globus_l_replica_catalog_arg_error(
		"Error opening password file for read." );
	    }

	    password = (char *) globus_libc_malloc(
		    GLOBUS_REPLICA_CATALOG_BUFSIZE * sizeof(char) );
	    fgets(password, GLOBUS_REPLICA_CATALOG_BUFSIZE, fd);
	    password[strlen(password)-1] = '\0';

	    if (strcmp(password_file,"-") != 0) {
		fclose(fd);
	    }
	}
    }

    if ( ( (action == arg_list_locations) ||
	    (action == arg_list_logicalfiles) ||
	    (action == arg_list_attributes) ||
	    (attributes_p == GLOBUS_TRUE) ) && (argc > 1) ) {
	attribute_names = (char **) globus_libc_malloc(
		argc * sizeof(char *) );
	for (i = 0; i < (argc - 1); i++) {
	    attribute_names[i] = globus_libc_strdup(argv[i+1]);
	}
	attribute_names[i] = GLOBUS_NULL;
    }

    if (action == arg_create) {
	switch (object) {
	    case arg_collection:
		if (argc > 1) {
		    filenames_file = globus_libc_strdup(argv[1]);
		}
		break;
	    case arg_location:
		if (argc > 1) {
		    location_url = globus_libc_strdup(argv[1]);
		    if (argc > 2) {
			filenames_file = globus_libc_strdup(argv[2]);
		    }
		} else {
		    globus_l_replica_catalog_arg_error(
		    "Specify a location URL." );
		}
		break;
	    case arg_logicalfile:
		if (argc > 1) {
		    size = globus_libc_strdup(argv[1]);
		} else {
		    globus_l_replica_catalog_arg_error(
		    "Specify a file size." );
		}
		break;
	    default:
		globus_l_replica_catalog_arg_error(
		"Error parsing arguments." );
		break;
	}
    }

    if (filenames_file != GLOBUS_NULL) {

	if (strcmp(filenames_file,"-") == 0) {
	    fd = stdin;
	} else if ((fd = fopen(filenames_file, "r")) == NULL) {
	    globus_l_replica_catalog_arg_error(
	    "Error opening filenames file for read." );
	}

	filenames = (char **) globus_libc_malloc(
		GLOBUS_REPLICA_CATALOG_BUFSIZE * sizeof(char *) );
	i = 0;
	while (1 != 2) {
	    filenames[i] = (char *) globus_libc_malloc(
		    GLOBUS_REPLICA_CATALOG_BUFSIZE * sizeof(char) );
	    if ( fgets( filenames[i], GLOBUS_REPLICA_CATALOG_BUFSIZE,
		    fd ) == NULL ) {
		break;
	    }
	    filenames[i][strlen(filenames[i])-1] = '\0';
	    i++;
	    if ((i % GLOBUS_REPLICA_CATALOG_BUFSIZE) == 0) {
		filenames = (char **) globus_libc_realloc( filenames,
			(i + GLOBUS_REPLICA_CATALOG_BUFSIZE)
			* sizeof(char *) );
	    }
	}
	filenames[i] = GLOBUS_NULL;

	if (strcmp(filenames_file,"-") != 0) {
	    fclose(fd);
	}
    }

#if 0
    if ( (action == arg_add_attribute)
	    || (action == arg_delete_attribute)
	    || (action == arg_add_filenames)
	    || (action == arg_delete_filenames)
	    || (action == arg_create)
	    || (action == arg_delete) ) {
	/* the same actions as the last check above */
    }
#endif

    /* TODO: add checking for extraneous arguments */
    /* TODO: handle objectclasses */
    /* TODO: handle non-cleartext authentication */

    /*
     * do actions
     */

    globus_replica_catalog_collection_handleattr_init(&handleattrs);
    globus_replica_catalog_collection_handleattr_set_authentication_mode(
	    &handleattrs,
	    GLOBUS_REPLICA_CATALOG_AUTHENTICATION_MODE_CLEARTEXT,
	    manager, password );
    if (!((object == arg_collection) && (action == arg_create))) {
	if ( ( err = globus_replica_catalog_collection_open( &handle,
		&handleattrs, collection_url ) ) != GLOBUS_SUCCESS ) {
	    globus_l_replica_catalog_action_error(err);
	}
    }

    switch (object) {

	/* act on a collection */
	case arg_collection:
	    switch (action) {
		case arg_create:
		    if ( ( err = globus_replica_catalog_collection_create(
			    &handle,
			    &handleattrs,
			    collection_url,
			    filenames,
			    GLOBUS_NULL ) ) != GLOBUS_SUCCESS ) {
			globus_l_replica_catalog_action_error(err);
		    }
		    break;
		case arg_delete:
		    if ( ( err = globus_replica_catalog_collection_delete(
			    &handle ) ) != GLOBUS_SUCCESS ) {
			globus_l_replica_catalog_action_error(err);
		    }
		    break;
		case arg_add_attribute:
		    if ( ( err =
			    globus_replica_catalog_collection_add_attribute(
				    &handle,
				    attribute_name,
				    attribute_value ) ) != GLOBUS_SUCCESS ) {
			globus_l_replica_catalog_action_error(err);
		    }
		    break;
		case arg_delete_attribute:
		    if ( ( err =
			    globus_replica_catalog_collection_delete_attribute(
				    &handle,
				    attribute_name,
				    attribute_value ) ) != GLOBUS_SUCCESS ) {
			globus_l_replica_catalog_action_error(err);
		    }
		    break;
		case arg_list_attributes:
		    globus_replica_catalog_attribute_set_init(
			    &attributes );
		    if ( ( err =
			    globus_replica_catalog_collection_list_attributes(
				    &handle,
				    attribute_names,
				    &attributes ) ) != GLOBUS_SUCCESS ) {
			globus_l_replica_catalog_action_error(err);
		    }
		    globus_l_replica_catalog_print_attributes(
			    attributes );
		    globus_replica_catalog_attribute_set_destroy(
			    &attributes );
		    break;
		case arg_add_filenames:
		    if ( ( err =
			    globus_replica_catalog_collection_add_filenames(
				    &handle,
				    filenames,
				    add_anyway_p ) ) != GLOBUS_SUCCESS ) {
			globus_l_replica_catalog_action_error(err);
		    }

		    break;
		case arg_delete_filenames:
		    if ( ( err =
			    globus_replica_catalog_collection_delete_filenames(
				    &handle,
				    filenames ) ) != GLOBUS_SUCCESS ) {
			globus_l_replica_catalog_action_error(err);
		    }
		    break;
		case arg_list_filenames:
		    if ( ( err =
			    globus_replica_catalog_collection_list_filenames(
				    &handle,
				    &filenames ) ) != GLOBUS_SUCCESS ) {
			globus_l_replica_catalog_action_error(err);
		    }
		    globus_l_replica_catalog_print_filenames(
			    filenames );
		    globus_replica_catalog_attribute_set_free_values(
			    filenames );
		    break;
		case arg_list_locations:
		    globus_replica_catalog_entry_set_init(
			    &entries );
		    if ( ( err =
			    globus_replica_catalog_collection_list_locations(
				    &handle,
				    attribute_names,
				    &entries ) ) != GLOBUS_SUCCESS ) {
			globus_l_replica_catalog_action_error(err);
		    }
		    globus_l_replica_catalog_print_entries(
			    entries );
		    globus_replica_catalog_entry_set_destroy(
			    &entries );
		    break;
		case arg_list_logicalfiles:
		    globus_replica_catalog_entry_set_init(
			    &entries );
		    if ( ( err =
			    globus_replica_catalog_collection_list_logicalfiles(
				    &handle,
				    attribute_names,
				    &entries ) ) != GLOBUS_SUCCESS ) {
			globus_l_replica_catalog_action_error(err);
		    }
		    globus_l_replica_catalog_print_entries(
			    entries );
		    globus_replica_catalog_entry_set_destroy(
			    &entries );
		    break;
		case arg_find_locations:
		    globus_replica_catalog_entry_set_init(
			    &entries );
		    if ( ( err =
			    globus_replica_catalog_collection_find_locations(
				    &handle,
				    filenames,
				    GLOBUS_TRUE,
				    attribute_names,
				    &entries ) ) != GLOBUS_SUCCESS ) {
			globus_l_replica_catalog_action_error(err);
		    }
		    globus_l_replica_catalog_print_entries(
			    entries );
		    globus_replica_catalog_entry_set_destroy(
			    &entries );
		    break;
		default:
		    globus_l_replica_catalog_arg_error(
		    "Error acting on collection arguments." );
		    break;
	    }
	    break;

	/* act on a location */
	case arg_location:
	    switch (action) {
		case arg_create:
		    if ( ( err = globus_replica_catalog_location_create(
			    &handle,
			    name,
			    location_url,
			    filenames,
			    GLOBUS_NULL ) ) != GLOBUS_SUCCESS ) {
			globus_l_replica_catalog_action_error(err);
		    }
		    break;
		case arg_delete:
		    if ( ( err = globus_replica_catalog_location_delete(
			    &handle,
			    name ) ) != GLOBUS_SUCCESS ) {
			globus_l_replica_catalog_action_error(err);
		    }
		    break;
		case arg_add_attribute:
		    if ( ( err =
			    globus_replica_catalog_location_add_attribute(
				    &handle,
				    name,
				    attribute_name,
			    attribute_value ) ) != GLOBUS_SUCCESS ) {
			globus_l_replica_catalog_action_error(err);
		    }
		    break;
		case arg_delete_attribute:
		    if ( ( err =
			    globus_replica_catalog_location_delete_attribute(
				    &handle,
				    name,
				    attribute_name,
				    attribute_value ) ) != GLOBUS_SUCCESS ) {
			globus_l_replica_catalog_action_error(err);
		    }
		    break;
		case arg_list_attributes:
		    globus_replica_catalog_attribute_set_init(
			    &attributes );
		    if ( ( err =
			    globus_replica_catalog_location_list_attributes(
				    &handle,
				    name,
				    attribute_names,
				    &attributes ) ) != GLOBUS_SUCCESS ) {
			globus_l_replica_catalog_action_error(err);
		    }
		    globus_l_replica_catalog_print_attributes(
			    attributes );
		    globus_replica_catalog_attribute_set_destroy(
			    &attributes );
		    break;
		case arg_add_filenames:
		    if ( ( err =
			    globus_replica_catalog_location_add_filenames(
				    &handle,
				    name,
				    filenames,
				    add_anyway_p ) ) != GLOBUS_SUCCESS ) {
			globus_l_replica_catalog_action_error(err);
		    }
		    break;
		case arg_delete_filenames:
		    if ( ( err =
			    globus_replica_catalog_location_delete_filenames(
				    &handle,
				    name,
				    filenames ) ) != GLOBUS_SUCCESS ) {
			globus_l_replica_catalog_action_error(err);
		    }
		    break;
		case arg_list_filenames:
		    if ( ( err =
			    globus_replica_catalog_location_list_filenames(
				    &handle,
				    name,
				    &filenames ) ) != GLOBUS_SUCCESS ) {
			globus_l_replica_catalog_action_error(err);
		    }
		    globus_l_replica_catalog_print_filenames(
			    filenames );
		    globus_replica_catalog_attribute_set_free_values(
			    filenames );
		    break;
		default:
		    globus_l_replica_catalog_arg_error(
		    "Error acting on location arguments." );
		    break;
	    }
	    break;

	/* act on a logicalfile */
	case arg_logicalfile:
	    switch (action) {
		case arg_create:
		    if ( ( err = globus_replica_catalog_logicalfile_create(
			    &handle,
			    name,
			    atoi(size),
			    GLOBUS_NULL ) ) != GLOBUS_SUCCESS ) {
			globus_l_replica_catalog_action_error(err);
		    }
		    break;
		case arg_delete:
		    if ( ( err = globus_replica_catalog_logicalfile_delete(
			    &handle,
			    name ) ) != GLOBUS_SUCCESS ) {
			globus_l_replica_catalog_action_error(err);
		    }
		    break;
		case arg_add_attribute:
		    if ( ( err =
			    globus_replica_catalog_logicalfile_add_attribute(
				    &handle,
				    name,
				    attribute_name,
				    attribute_value ) ) != GLOBUS_SUCCESS ) {
			globus_l_replica_catalog_action_error(err);
		    }
		    break;
		case arg_delete_attribute:
		    if ( ( err =
			    globus_replica_catalog_logicalfile_delete_attribute(
				    &handle,
				    name,
				    attribute_name,
				    attribute_value ) ) != GLOBUS_SUCCESS ) {
			globus_l_replica_catalog_action_error(err);
		    }
		    break;
		case arg_list_attributes:
		    globus_replica_catalog_attribute_set_init(
			    &attributes );
		    if ( ( err =
			    globus_replica_catalog_logicalfile_list_attributes(
				    &handle,
				    name,
				    attribute_names,
				    &attributes ) ) != GLOBUS_SUCCESS ) {
			globus_l_replica_catalog_action_error(err);
		    }
		    globus_l_replica_catalog_print_attributes(
			    attributes );
		    globus_replica_catalog_attribute_set_destroy(
			    &attributes );
		    break;
		default:
		    globus_l_replica_catalog_arg_error(
		    "Error acting on logical file arguments." );
		    break;
	    }
	    break;

	default:
	    if (action == arg_list_collections) {
		globus_replica_catalog_entry_set_init(
			&entries );
		if ( ( err =
			globus_replica_catalog_list_collections(
				&handle,
				GLOBUS_NULL,
				&entries ) ) != GLOBUS_SUCCESS ) {
		    globus_l_replica_catalog_action_error(err);
		}
		globus_l_replica_catalog_print_entry_dns(
			entries );
		globus_replica_catalog_entry_set_destroy(
			&entries );
	    } else {
		globus_l_replica_catalog_arg_error(
		"Error acting on arguments." );
	    }
	    break;
    }

    /*
     * clean up
     */

    if (!((object == arg_collection) && (action == arg_delete))) {
	globus_replica_catalog_collection_close(&handle);
    }

    globus_replica_catalog_collection_handleattr_destroy(&handleattrs);

    globus_module_deactivate_all();

    return 0;
}

