/* $Id: globus_replica_management.c,v 1.35 2002/04/18 16:57:44 smartin Exp $ */

#ifndef GLOBUS_DONT_DOCUMENT_INTERNAL
/**
 * @file
 *
 * Globus Replica Management C API Source File
 *
 * This file defines types, variables, and functions that manage the
 * placement of files on remote storage systems, and that manage the
 * information about these files and their locations in a replica
 * catalog.
 */
#endif /* GLOBUS_DONT_DOCUMENT_INTERNAL */

#include "globus_replica_management.h"
#include "globus_i_replica_management.h"
#include "globus_i_replica_catalog.h"
#include "globus_replica_management_progress_plugin.h"
#include "version.h"

/* Local variable declarations */

static globus_list_t * globus_l_replica_management_handle_list = GLOBUS_NULL;
static globus_mutex_t  globus_l_replica_management_handle_list_mutex;

/* Local function declarations */

static globus_result_t
globus_l_replica_management_add_cache_entry(
    globus_replica_management_handle_t *                handle,
    char *                                              collection_url,
    globus_replica_catalog_collection_handleattr_t *    collection_attrs);

static globus_result_t
globus_l_replica_management_get_catalog_handle(
    globus_replica_management_handle_t *                handle,
    char *                                              collection_url,
    globus_replica_catalog_collection_handleattr_t *    collection_attrs);

static globus_result_t
globus_l_replica_management_release_catalog_handle(
    globus_replica_management_handle_t *    handle);

static globus_result_t
globus_l_replica_management_destroy_catalog_handle(
    globus_replica_management_handle_t *    handle);

static globus_result_t
globus_l_replica_management_flush_catalog_cache(
    globus_replica_management_handle_t *    handle);

static int
globus_l_replica_management_catalog_handle_compare(
    void *                              entry1,
    void *                              entry2);

static void
globus_l_replica_management_ftp_exists_cb(
    void *                                  user_arg,
    globus_ftp_client_handle_t *            handle,
    globus_object_t *                       err);

static void
globus_l_replica_management_ftp_cmd_cb(
    void *                                  user_arg,
    globus_ftp_client_handle_t *            handle,
    globus_object_t *                       err);

static void
globus_l_gass_copy_callback(
    void *                                  callback_arg,
    globus_gass_copy_handle_t *             handle,
    globus_object_t *                       err);

static globus_result_t
globus_l_replica_management_file_move(
    globus_replica_management_handle_t *              handle,
    char *                                            source_url,
    globus_ftp_client_operationattr_t *               source_attrs,
    char *                                            dest_collection_url,
    globus_replica_catalog_collection_handleattr_t *  dest_collection_attrs,
    char *                                            dest_location_name,
    globus_ftp_client_operationattr_t *               dest_location_attrs,
    char *                                            dest_filename,
    globus_replica_management_overwrite_behavior_t    overwrite_behavior,
    globus_replica_management_preregister_function_t  preregister,
    void *                                            preregister_arg,
    globus_bool_t                                     copy);

typedef enum globus_l_replica_management_restart_state_e
{
    GLOBUS_L_REPLICA_MANAGEMENT_PHYSICAL_DELETE_START,
    GLOBUS_L_REPLICA_MANAGEMENT_LOGICAL_DELETE_START,
    GLOBUS_L_REPLICA_MANAGEMENT_NEED_PHYSICAL_DELETE,
    GLOBUS_L_REPLICA_MANAGEMENT_SAFE_START,
    GLOBUS_L_REPLICA_MANAGEMENT_UNSAFE_START,
    GLOBUS_L_REPLICA_MANAGEMENT_NEED_FILE_MOVE,
    GLOBUS_L_REPLICA_MANAGEMENT_NEED_PREREGISTER,
    GLOBUS_L_REPLICA_MANAGEMENT_NEED_COLLECTION_ADD,
    GLOBUS_L_REPLICA_MANAGEMENT_NEED_LOCATION_ADD
}
globus_l_replica_management_restart_state_t;

static globus_l_replica_management_restart_state_t
globus_l_replica_management_get_state(
    char *                              state);

static globus_result_t
globus_l_replica_management_create_dir(
    globus_replica_management_handle_t *    handle,
    globus_ftp_client_operationattr_t *     ftp_attrs,
    char *				    root_dir_url);

static int
globus_l_replica_management_activate(void);

static int
globus_l_replica_management_deactivate(void);

#ifndef GLOBUS_DONT_DOCUMENT_INTERNAL

/**
 * Module descriptor.
 */
globus_module_descriptor_t		globus_i_replica_management_module =
{
    "globus_replica_management",
    globus_l_replica_management_activate,
    globus_l_replica_management_deactivate,
    GLOBUS_NULL,
    GLOBUS_NULL,
    &local_version
};

/**
 * Module activation callback.
 */
static int
globus_l_replica_management_activate(void)
{
    globus_module_activate(GLOBUS_FTP_CLIENT_MODULE);
    globus_module_activate(GLOBUS_REPLICA_CATALOG_MODULE);
    globus_module_activate(GLOBUS_GASS_COPY_MODULE);
    globus_mutex_init(
	&(globus_l_replica_management_handle_list_mutex), GLOBUS_NULL);
    return GLOBUS_SUCCESS;
}

/**
 * Module deactivation callback.
 */
static int
globus_l_replica_management_deactivate(void)
{
    globus_replica_management_handle_t *    handle;

    globus_mutex_lock(&globus_l_replica_management_handle_list_mutex);
    {
	while(!globus_list_empty(globus_l_replica_management_handle_list))
	{
	    handle = (globus_replica_management_handle_t *)
		globus_list_remove(&globus_l_replica_management_handle_list,
				   globus_l_replica_management_handle_list);

	    globus_mutex_destroy(&handle->mutex);
	    	
	    globus_ftp_client_handle_destroy(&handle->ftp_handle);
	    
	    globus_replica_management_handleattr_destroy(&(handle->attrs));
		
	    globus_i_replica_management_progress_plugin_destroy(
		&handle->plugin); 
		
	    handle->cache_entry = GLOBUS_NULL;
		
	    if(handle->attrs.caching == GLOBUS_TRUE)
	    {
		globus_l_replica_management_flush_catalog_cache(handle);
	    }

	    handle->replica_handle = GLOBUS_NULL;
	    
	    globus_gass_copy_handle_destroy(&handle->gass_copy_handle);
	}
    }
    globus_mutex_unlock(&globus_l_replica_management_handle_list_mutex);

    globus_mutex_destroy(&globus_l_replica_management_handle_list_mutex);
    
    globus_module_deactivate(GLOBUS_FTP_CLIENT_MODULE);
    globus_module_deactivate(GLOBUS_REPLICA_CATALOG_MODULE);
    globus_module_deactivate(GLOBUS_GASS_COPY_MODULE);
    
    return GLOBUS_SUCCESS;
}

#endif /* GLOBUS_DONT_DOCUMENT_INTERNAL */

/* Session Management */

globus_result_t
globus_replica_management_handleattr_init(
	globus_replica_management_handleattr_t *    attrs)
{
    globus_result_t                     rc;
    static char * myname = "globus_replica_management_handleattr_init";

    if(attrs == GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
		"NULL handle attributes",
		myname);
	goto return_result;
    }

    /* XXX: should this be here? */
    attrs->third_party_transfer = GLOBUS_TRUE;

    attrs->caching = GLOBUS_TRUE; /* set caching to true */
    attrs->ftp_update_frequency = 20; /* update every 100 seconds */
    rc = globus_ftp_client_handleattr_init(&attrs->ftp_handle_attrs);
    
 return_result:
    return rc;
}

globus_result_t
globus_replica_management_handleattr_destroy(
    globus_replica_management_handleattr_t * attrs)
{
    globus_result_t                     rc;
    static char * myname = "globus_replica_management_handleattr_destroy";

    if(attrs == GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
		"NULL handle attributes",
		myname);
	goto return_result;
    }

    rc = globus_ftp_client_handleattr_destroy(&attrs->ftp_handle_attrs);

    if(rc != GLOBUS_SUCCESS)
    {
	goto return_result;
    }
    
    return GLOBUS_SUCCESS;
    
 return_result:
    return rc;
}

globus_result_t
globus_replica_management_handleattr_copy(
    globus_replica_management_handleattr_t *    dest,
    globus_replica_management_handleattr_t *    src)
{
    globus_result_t                     rc;
    static char * myname = "globus_replica_management_handleattr_copy";
    
    if(dest == GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
		"NULL handle attributes",
		myname);
	goto return_result;
    }

    if(src == GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "NULL handle attributes",
	    myname);
	goto return_result;
    }

    dest->caching = src->caching;
    dest->ftp_update_frequency = src->ftp_update_frequency;

    /* XXX: should this be here? */
    dest->third_party_transfer = src->third_party_transfer;

    rc = globus_ftp_client_handleattr_copy(&dest->ftp_handle_attrs,
					   &src->ftp_handle_attrs);

    if(rc != GLOBUS_SUCCESS)
    {
	goto return_result;
    }

    return GLOBUS_SUCCESS;

 return_result:
    return rc;
}

globus_result_t
globus_replica_management_handleattr_set_caching(
    globus_replica_management_handleattr_t *    attrs,
    globus_bool_t                               caching)
{
    globus_result_t                     rc;
    static char * myname = "globus_replica_management_handleattr_set_caching";

    if(attrs == GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
		"NULL handle attributes",
		myname);
	goto return_result;
    }
    
    attrs->caching = caching;

    rc = GLOBUS_SUCCESS;
    
 return_result:
    return rc;
}

globus_result_t
globus_replica_management_handleattr_get_caching(
    globus_replica_management_handleattr_t *    attrs,
    globus_bool_t *                             caching)
{
    globus_result_t                     rc;
    static char * myname = "globus_replica_management_handleattr_get_caching";

    if(attrs == GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
		"NULL handle attributes",
		myname);
	goto return_result;
    }

    *caching = attrs->caching;

    rc = GLOBUS_SUCCESS;
    
 return_result:
    return rc;
}

globus_result_t
globus_replica_management_handleattr_set_ftp_update_frequency(
    globus_replica_management_handleattr_t *    attrs,
    unsigned int                                frequency)
{
    globus_result_t                     rc;
    static char * myname = "globus_replica_management_handleattr_set_ftp_update_frequency";

    if(attrs == GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
		"NULL handle attributes",
		myname);
	goto return_result;
    }

    attrs->ftp_update_frequency = frequency;

    rc = GLOBUS_SUCCESS;
    
 return_result:
    return rc;
}

globus_result_t
globus_replica_management_handleattr_get_ftp_update_frequency(
    globus_replica_management_handleattr_t *    attrs,
    unsigned int *                              frequency)
{
    globus_result_t                     rc;
    static char * myname = "globus_replica_management_handleattr_get_ftp_update_frequency";

    if(attrs == GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
		"NULL handle attributes",
		myname);
	goto return_result;
    }

    *frequency = attrs->ftp_update_frequency;

    rc = GLOBUS_SUCCESS;
    
 return_result:
    return rc;
}

globus_result_t
globus_replica_management_handleattr_set_ftp_handleattr(
    globus_replica_management_handleattr_t *    attrs,
    globus_ftp_client_handleattr_t *            ftp_handle_attrs)
{
    globus_result_t                     rc;
    static char * myname = "globus_replica_management_handleattr_set_ftp_handleattr";

    if(attrs == GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
		"NULL handle attributes",
		myname);
	goto return_result;
    }

    if(ftp_handle_attrs == GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
		"NULL ftp handle attributes",
		myname);
	goto return_result;
    }
    
    rc = globus_ftp_client_handleattr_destroy(&attrs->ftp_handle_attrs);

    if(rc != GLOBUS_SUCCESS)
    {
	goto return_result;
    }
    
    rc = globus_ftp_client_handleattr_copy(&attrs->ftp_handle_attrs,
					   ftp_handle_attrs);
    
 return_result:
    return rc;
}

globus_result_t
globus_replica_management_handleattr_get_ftp_handleattr(
    globus_replica_management_handleattr_t *    attrs,
    globus_ftp_client_handleattr_t *            ftp_handle_attrs)
{
    globus_result_t                     rc;
    static char * myname = "globus_replica_management_handleattr_get_ftp_handleattr";

    if(attrs == GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
		"NULL handle attributes",
		myname);
	goto return_result;
    }

    if(ftp_handle_attrs == GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
		"NULL ftp handle attributes",
		myname);
	goto return_result;
    }
    
    rc = globus_ftp_client_handleattr_copy(ftp_handle_attrs,
					   &attrs->ftp_handle_attrs);
 return_result:
    return rc;
}

/* XXX: should this act on the handleattrs or set it directly in the
 * handle like gass_copy? */
globus_result_t
globus_replica_management_handleattr_set_third_party_transfer(
    globus_replica_management_handleattr_t *    attrs,
    globus_bool_t                               third_party_transfer)
{
    if (attrs != GLOBUS_NULL)
    {
	attrs->third_party_transfer = third_party_transfer;
    }
    return GLOBUS_SUCCESS;
}

/* XXX: should this act on the handleattrs or get it directly in the
 * handle like gass_copy? */
globus_result_t
globus_replica_management_handleattr_get_third_party_transfer(
    globus_replica_management_handleattr_t *    attrs,
    globus_bool_t *                             third_party_transfer)
{
    if (attrs != GLOBUS_NULL)
    {
	*third_party_transfer = attrs->third_party_transfer;
    }
    return GLOBUS_SUCCESS;
}

globus_result_t
globus_replica_management_handle_init(
	globus_replica_management_handle_t *        handle,
	globus_replica_management_handleattr_t *    attrs )
{
    globus_result_t                     rc;
    globus_gass_copy_handleattr_t  	gass_copy_handleattrs;
    static char * myname = "globus_replica_management_handle_init";

    if(handle == GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "NULL handle",
	    myname);
	goto return_result;
    }
    
    if(attrs == GLOBUS_NULL)
    {
	rc = globus_replica_management_handleattr_init(&handle->attrs);
    }
    else
    {
	rc = globus_replica_management_handleattr_copy(&(handle->attrs),
						       attrs);
    }

    if(rc != GLOBUS_SUCCESS)
    {
	goto return_result;
    }

    globus_mutex_init(&handle->mutex, GLOBUS_NULL);

    handle->replica_cache = GLOBUS_NULL;

    handle->cache_entry = GLOBUS_NULL;
    
    handle->replica_handle = GLOBUS_NULL;

    rc = globus_i_replica_management_plugin_data_init(handle);
    
    if(rc != GLOBUS_SUCCESS)
    {
	goto destroy_attrs;
    }
    
    rc = globus_i_replica_management_progress_plugin_init(
	&handle->plugin,
	&handle->plugin_data);

    if(rc != GLOBUS_SUCCESS)
    {
	goto destroy_attrs;
    }

    rc = globus_ftp_client_handleattr_add_plugin(&attrs->ftp_handle_attrs,
						 &handle->plugin);

    if(rc != GLOBUS_SUCCESS)
    {
	goto destroy_plugin;
    }
  
    rc = globus_ftp_client_handle_init(&handle->ftp_handle,
				       &attrs->ftp_handle_attrs);
    
    if(rc != GLOBUS_SUCCESS)
    {
	goto destroy_plugin;
    }

    rc = globus_gass_copy_handleattr_init(&gass_copy_handleattrs);
    
    if (rc != GLOBUS_SUCCESS)
    {
	goto destroy_plugin;
    }

    rc = globus_gass_copy_handleattr_set_ftp_attr(&gass_copy_handleattrs,
						  &attrs->ftp_handle_attrs );
    if (rc != GLOBUS_SUCCESS)
    {
	goto destroy_gass_copy_handleattrs;
    }

    rc = globus_gass_copy_handle_init(&handle->gass_copy_handle,
					&gass_copy_handleattrs);
    if (rc != GLOBUS_SUCCESS)
    {
	goto destroy_gass_copy_handleattrs;
    }

    if (handle->attrs.third_party_transfer == GLOBUS_FALSE)
    {
	rc = globus_gass_copy_set_no_third_party_transfers(
	    &handle->gass_copy_handle,
	    GLOBUS_TRUE);
	if (rc != GLOBUS_SUCCESS)
	{
	    goto destroy_gass_copy_handle;
	}
    }
    else
    {
	rc = globus_gass_copy_set_no_third_party_transfers(
	    &handle->gass_copy_handle,
	    GLOBUS_FALSE);
	if (rc != GLOBUS_SUCCESS)
	{
	    goto destroy_gass_copy_handle;
	}
    }

    globus_mutex_lock(&globus_l_replica_management_handle_list_mutex);
    {
	globus_list_insert(&globus_l_replica_management_handle_list,
			   (void *) handle);
	handle->list_elem = globus_l_replica_management_handle_list;
    }
    globus_mutex_unlock(&globus_l_replica_management_handle_list_mutex);

    globus_gass_copy_handleattr_destroy(&gass_copy_handleattrs);
    
    return GLOBUS_SUCCESS;

 destroy_gass_copy_handle:
    globus_gass_copy_handle_destroy(&handle->gass_copy_handle);
 destroy_gass_copy_handleattrs:
    globus_gass_copy_handleattr_destroy(&gass_copy_handleattrs);
 destroy_plugin:
    globus_i_replica_management_progress_plugin_destroy(&handle->plugin);
 destroy_attrs:
    globus_replica_management_handleattr_destroy(&handle->attrs);
 return_result:
    return rc;
}

globus_result_t
globus_replica_management_handle_destroy(
	globus_replica_management_handle_t *    handle)
{
    globus_result_t                     rc;
    void *                              result;
    static char * myname = "globus_replica_management_handle_destroy";

    if(handle == GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "NULL handle",
	    myname);
	goto return_result;
    }

    rc = GLOBUS_SUCCESS;
    
    globus_mutex_lock(&globus_l_replica_management_handle_list_mutex);
    {
	result = globus_list_remove(&globus_l_replica_management_handle_list,
				    handle->list_elem);
    }
    globus_mutex_unlock(&globus_l_replica_management_handle_list_mutex);

    if(result != GLOBUS_NULL)
    {
	globus_mutex_destroy(&handle->mutex);

	globus_ftp_client_handle_destroy(&handle->ftp_handle);
	
	globus_replica_management_handleattr_destroy(&(handle->attrs));
	
	globus_i_replica_management_progress_plugin_destroy(&handle->plugin); 
	
	handle->cache_entry = GLOBUS_NULL;
	
	if(handle->attrs.caching == GLOBUS_TRUE)
	{
	    rc = globus_l_replica_management_flush_catalog_cache(handle);
	}

	handle->replica_handle = GLOBUS_NULL;

	globus_gass_copy_handle_destroy(&handle->gass_copy_handle);
    }

 return_result:
    return rc;
}

/* Catalog Creation */

globus_result_t
globus_replica_management_collection_create(
    globus_replica_management_handle_t *              handle,
    char *                                            collection_url,
    globus_replica_catalog_collection_handleattr_t *  collection_attrs)
{
    globus_result_t                                     rc;
    globus_replica_catalog_collection_handleattr_t      attrs;
    static char * myname = "globus_replica_management_collection_create";

    if(handle == GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "NULL handle",
	    myname);
	goto return_result;
    }
    
    if(collection_url == GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "NULL URL",
	    myname);
	goto return_result;
    }

    if(handle->replica_handle != GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "Busy handle",
	    myname);
	goto return_result;
    }

    
    if(collection_attrs == GLOBUS_NULL)
    {
	rc = globus_replica_catalog_collection_handleattr_init(&attrs);
    }
    else
    {
	rc = globus_replica_catalog_collection_handleattr_copy(
	    &attrs,
	    collection_attrs);
    }
    
    if(rc != GLOBUS_SUCCESS)
    {
	goto return_result;
    }

    handle->replica_handle = globus_libc_malloc(
	sizeof(globus_replica_catalog_collection_handle_t));

    if(handle->replica_handle == GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "Could not allocate internal data structure",
	    myname);
	goto destroy_collection_attrs;
    }

    rc = globus_replica_catalog_collection_create(handle->replica_handle,
						  &attrs,
						  collection_url,
						  GLOBUS_NULL,
						  GLOBUS_NULL);
    if(rc != GLOBUS_SUCCESS)
    {
	goto result_free_handle;
    }

    if(handle->attrs.caching == GLOBUS_TRUE)
    {
	rc = globus_l_replica_management_add_cache_entry(handle,
							 collection_url,
							 &attrs);
	if(rc != GLOBUS_SUCCESS)
	{
	    goto result_destroy_handle;
	}
	
	handle->replica_handle = GLOBUS_NULL;
	handle->cache_entry = GLOBUS_NULL;
	globus_replica_catalog_collection_handleattr_destroy(&attrs);
	return GLOBUS_SUCCESS;
    }
    else
    {
	/* set return value and continue */
	rc =  GLOBUS_SUCCESS;
    }
    
 result_destroy_handle:
    globus_replica_catalog_collection_close(handle->replica_handle);
 result_free_handle:
    globus_libc_free(handle->replica_handle);
    handle->replica_handle = GLOBUS_NULL;
 destroy_collection_attrs:
    globus_replica_catalog_collection_handleattr_destroy(&attrs);
 return_result:
    return rc;
}

globus_result_t
globus_replica_management_collection_delete(
    globus_replica_management_handle_t *              handle,
    char *                                            collection_url,
    globus_replica_catalog_collection_handleattr_t *  collection_attrs)
{
    globus_result_t                                     rc;
    globus_replica_catalog_collection_handleattr_t      attrs;
    globus_l_replica_management_cache_entry_t *         entry;
    static char * myname = "globus_replica_management_collection_delete";

    if(handle == GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "NULL handle",
	    myname);
	goto return_result;
    }
    
    if(collection_url == GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "NULL URL",
	    myname);
	goto return_result;
    }

    if(handle->replica_handle != GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "Busy handle",
	    myname);
	goto return_result;
    }

    
    if(collection_attrs == GLOBUS_NULL)
    {
	rc = globus_replica_catalog_collection_handleattr_init(&attrs);
    }
    else
    {
	rc = globus_replica_catalog_collection_handleattr_copy(
	    &attrs,
	    collection_attrs);
    }
    
    if(rc != GLOBUS_SUCCESS)
    {
	goto return_result;
    }

    rc = globus_l_replica_management_get_catalog_handle(handle,
							collection_url,
							&attrs);
    if(rc != GLOBUS_SUCCESS)
    {
	goto destroy_collection_attrs;
    }

    rc = globus_replica_catalog_collection_delete(handle->replica_handle);

    if(rc != GLOBUS_SUCCESS)
    {
	globus_l_replica_management_destroy_catalog_handle(handle);
	goto destroy_collection_attrs;
    }
    
    if(handle->attrs.caching == GLOBUS_TRUE)
    {
	entry = (globus_l_replica_management_cache_entry_t *)
	    globus_list_remove(&handle->replica_cache,
			       handle->cache_entry);
	globus_replica_catalog_collection_handleattr_destroy(&entry->attrs);
	globus_libc_free(entry->url);
	globus_libc_free(entry);
	handle->cache_entry = GLOBUS_NULL;
    }
    else
    {
	globus_libc_free(handle->replica_handle);
	handle->replica_handle = GLOBUS_NULL;
    }
    
 destroy_collection_attrs:
    globus_replica_catalog_collection_handleattr_destroy(&attrs);
 return_result:
    return rc;
}

globus_result_t
globus_replica_management_location_create(
    globus_replica_management_handle_t *              handle,
    char *                                            collection_url,
    globus_replica_catalog_collection_handleattr_t *  collection_attrs,
    char *                                            location_name,
    globus_ftp_client_operationattr_t *               location_attrs,
    char *                                            root_dir_url)
{
    globus_result_t                                     rc;
    globus_replica_catalog_collection_handleattr_t      attrs;
    globus_ftp_client_operationattr_t                   ftp_attrs;
    static char * myname = "globus_replica_management_location_create";

    if(handle == GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "NULL handle",
	    myname);
	goto return_result;
    }

    if(handle->replica_handle != GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "Busy handle",
	    myname);
	goto return_result;
    }

    if(collection_url == GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "NULL URL",
	    myname);
	goto return_result;
    }

    if(location_name == GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "NULL location name",
	    myname);
	goto return_result;
    }

    if(root_dir_url == GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "NULL URL",
	    myname);
	goto return_result;
    }

    
    if(collection_attrs == GLOBUS_NULL)
    {
	rc = globus_replica_catalog_collection_handleattr_init(&attrs);
    }
    else
    {
	rc = globus_replica_catalog_collection_handleattr_copy(
	    &attrs,
	    collection_attrs);
    }
    
    if(rc != GLOBUS_SUCCESS)
    {
	goto return_result;
    }

    if(location_attrs == GLOBUS_NULL)
    {
	rc = globus_ftp_client_operationattr_init(&ftp_attrs);
    }
    else
    {
	rc = globus_ftp_client_operationattr_copy(&ftp_attrs,
						  location_attrs);
    }
    
    if(rc != GLOBUS_SUCCESS)
    {
	goto destroy_collection_attrs;
    }


    rc = globus_l_replica_management_get_catalog_handle(handle,
							collection_url,
							&attrs);
    if(rc != GLOBUS_SUCCESS)
    {
	goto destroy_ftp_attrs;
    }

    rc = globus_l_replica_management_create_dir(handle,
						&ftp_attrs,
						root_dir_url);
    if(rc != GLOBUS_SUCCESS)
    {
	globus_l_replica_management_release_catalog_handle(handle);
	goto destroy_ftp_attrs;
    }
    
    rc = globus_replica_catalog_location_create(handle->replica_handle,
						location_name,
						root_dir_url,
						GLOBUS_NULL,
						GLOBUS_NULL);
    if(rc != GLOBUS_SUCCESS)
    {
	globus_l_replica_management_destroy_catalog_handle(handle);
	goto destroy_ftp_attrs;
    }

    rc = globus_l_replica_management_release_catalog_handle(handle);

 destroy_ftp_attrs:
    globus_ftp_client_operationattr_destroy(&ftp_attrs);
 destroy_collection_attrs:
    globus_replica_catalog_collection_handleattr_destroy(&attrs);
 return_result:
    return rc;
}

globus_result_t
globus_replica_management_location_delete(
    globus_replica_management_handle_t *              handle,
    char *                                            collection_url,
    globus_replica_catalog_collection_handleattr_t *  collection_attrs,
    char *                                            location_name)
{
    globus_result_t                                     rc;
    globus_replica_catalog_collection_handleattr_t      attrs;
    static char * myname = "globus_replica_management_location_delete";

    if(handle == GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "NULL handle",
	    myname);
	goto return_result;
    }

    if(handle->replica_handle != GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "Busy handle",
	    myname);
	goto return_result;
    }

    if(collection_url == GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "NULL URL",
	    myname);
	goto return_result;
    }

    if(location_name == GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "NULL location name",
	    myname);
	goto return_result;
    }

    if(collection_attrs == GLOBUS_NULL)
    {
	rc = globus_replica_catalog_collection_handleattr_init(&attrs);
    }
    else
    {
	rc = globus_replica_catalog_collection_handleattr_copy(
	    &attrs,
	    collection_attrs);
    }
    
    if(rc != GLOBUS_SUCCESS)
    {
	goto return_result;
    }


    rc = globus_l_replica_management_get_catalog_handle(handle,
							collection_url,
							&attrs);
    if(rc != GLOBUS_SUCCESS)
    {
	goto destroy_collection_attrs;
    }
    
    rc = globus_replica_catalog_location_delete(handle->replica_handle,
						location_name);

    if(rc != GLOBUS_SUCCESS)
    {
	goto close_handle;
    }

    rc = globus_l_replica_management_release_catalog_handle(handle);
    
    goto destroy_collection_attrs;
    
 close_handle:
    globus_l_replica_management_destroy_catalog_handle(handle);
 destroy_collection_attrs:
    globus_replica_catalog_collection_handleattr_destroy(&attrs);
 return_result:
    return rc;
}

globus_result_t
globus_replica_management_file_register(
    globus_replica_management_handle_t *              handle,
    char *                                            collection_url,
    globus_replica_catalog_collection_handleattr_t *  collection_attrs,
    char *                                            location_name,
    globus_ftp_client_operationattr_t *               location_attrs,
    char *                                            filename)
{
    globus_result_t                                     rc;
    globus_replica_catalog_collection_handleattr_t      coll_attrs;
    globus_ftp_client_operationattr_t                   ftp_attrs;
    globus_i_replica_management_monitor_t               monitor;
    globus_byte_t                                       buffer[GLOBUS_L_REPLICA_MANAGEMENT_BUFFER_SIZE];
    char *                                              url;
    char *                                              filenames[2];
    static char * myname = "globus_replica_management_file_register";

    if(handle == GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "NULL handle",
	    myname);
	goto return_result;
    }

    if(handle->replica_handle != GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "Busy handle",
	    myname);
	goto return_result;
    }
    
    if(collection_url == GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "NULL URL",
	    myname);
	goto return_result;
    }

    if(location_name == GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "NULL location name",
	    myname);
	goto return_result;
    }

    if(filename == GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "NULL file name",
	    myname);
	goto return_result;
    }

    
    if(collection_attrs == GLOBUS_NULL)
    {
	rc = globus_replica_catalog_collection_handleattr_init(&coll_attrs);
    }
    else
    {
	rc = globus_replica_catalog_collection_handleattr_copy(
	    &coll_attrs,
	    collection_attrs);
    }
    
    if(rc != GLOBUS_SUCCESS)
    {
	goto return_result;
    }


    if(location_attrs == GLOBUS_NULL)
    {
	rc = globus_ftp_client_operationattr_init(&ftp_attrs);
    }
    else
    {
	rc = globus_ftp_client_operationattr_copy(&ftp_attrs,
						  location_attrs);
    }
    
    if(rc != GLOBUS_SUCCESS)
    {
	goto destroy_collection_attrs;
    }

    
    rc = globus_i_replica_management_monitor_init(&monitor);

    if(rc != GLOBUS_SUCCESS)
    {
	goto destroy_ftp_attrs;
    }

    rc = globus_l_replica_management_get_catalog_handle(handle,
							collection_url,
							&coll_attrs);
    if(rc != GLOBUS_SUCCESS)
    {
	goto destroy_monitor;
    }

    rc = globus_i_replica_catalog_location_get_url(
	handle->replica_handle,
	location_name,
	filename,
	&url);

    if(rc != GLOBUS_SUCCESS)
    {
	globus_l_replica_management_destroy_catalog_handle(handle);
	goto destroy_monitor;
    }
	
    rc = globus_ftp_client_exists(&handle->ftp_handle,
				url,
				&ftp_attrs,
				globus_l_replica_management_ftp_exists_cb,
				(void *) &monitor);
    if(rc != GLOBUS_SUCCESS)
    {
	globus_l_replica_management_release_catalog_handle(handle);
	goto destroy_url;
    }

    globus_mutex_lock(&monitor.mutex);
    {
	while(monitor.done == GLOBUS_FALSE)
	{
	    globus_cond_wait(&monitor.cond, &monitor.mutex);
	}
    }
    globus_mutex_unlock(&monitor.mutex);

    if(monitor.result != GLOBUS_SUCCESS)
    {
	rc = monitor.result;
	globus_l_replica_management_release_catalog_handle(handle);
	goto destroy_url;
    }


    filenames[0] = filename;
    filenames[1] = GLOBUS_NULL;
    
    rc = globus_replica_catalog_collection_add_filenames(
	handle->replica_handle,
	filenames,
	GLOBUS_TRUE);

    if(rc != GLOBUS_SUCCESS)
    {
	globus_l_replica_management_destroy_catalog_handle(handle);
	goto destroy_url;
    }

    rc = globus_replica_catalog_location_add_filenames(
	handle->replica_handle,
	location_name,
	filenames,
	GLOBUS_TRUE);

    if(rc != GLOBUS_SUCCESS)
    {
	globus_l_replica_management_destroy_catalog_handle(handle);
	goto destroy_url;
    }

    rc = globus_l_replica_management_release_catalog_handle(handle);

 destroy_url:
    globus_libc_free(url);
 destroy_monitor:
    globus_i_replica_management_monitor_destroy(&monitor);
 destroy_ftp_attrs:
    globus_ftp_client_operationattr_destroy(&ftp_attrs);
 destroy_collection_attrs:
    globus_replica_catalog_collection_handleattr_destroy(&coll_attrs);
 return_result:
    return rc;
}

globus_result_t
globus_replica_management_file_publish(
    globus_replica_management_handle_t *              handle,
    char *                                            source_url,
    globus_ftp_client_operationattr_t *               source_attrs,
    char *                                            dest_collection_url,
    globus_replica_catalog_collection_handleattr_t *  dest_collection_attrs,
    char *                                            dest_location_name,
    globus_ftp_client_operationattr_t *               dest_location_attrs,
    char *                                            dest_filename,
    globus_replica_management_overwrite_behavior_t    overwrite_behavior,
    globus_replica_management_preregister_function_t  preregister,
    void *                                            preregister_arg)
{
    globus_result_t                                     rc;
    globus_replica_catalog_collection_handleattr_t      collection_attrs;
    globus_ftp_client_operationattr_t                   src_ftp_attrs;
    globus_ftp_client_operationattr_t                   dest_ftp_attrs;
    static char * myname = "globus_replica_management_file_publish";

    if(handle == GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "NULL handle",
	    myname);
	goto return_result;
    }

    if(handle->replica_handle != GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "Busy handle",
	    myname);
	goto return_result;
    }

    if(source_url == GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "NULL URL",
	    myname);
	goto return_result;
    }

    if(dest_collection_url == GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "NULL URL",
	    myname);
	goto return_result;
    }

    
    if(dest_location_name == GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "NULL location name",
	    myname);
	goto return_result;
    }

    if(dest_filename == GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "NULL file name",
	    myname);
	goto return_result;
    }

    if(overwrite_behavior != GLOBUS_REPLICA_MANAGEMENT_OVERWRITE_DISALLOW &&
       overwrite_behavior != GLOBUS_REPLICA_MANAGEMENT_OVERWRITE_SAFE     &&
       overwrite_behavior != GLOBUS_REPLICA_MANAGEMENT_OVERWRITE_UNSAFE)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "invalid overwrite behavior",
	    myname);
	goto return_result;
    }
    
    if(dest_collection_attrs == GLOBUS_NULL)
    {
	rc = globus_replica_catalog_collection_handleattr_init(
	    &collection_attrs);
    }
    else
    {
	rc = globus_replica_catalog_collection_handleattr_copy(
	    &collection_attrs,
	    dest_collection_attrs);
    }
    
    if(rc != GLOBUS_SUCCESS)
    {
	goto return_result;
    }

    if(source_attrs == GLOBUS_NULL)
    {
	rc = globus_ftp_client_operationattr_init(&src_ftp_attrs);
    }
    else
    {
	rc = globus_ftp_client_operationattr_copy(&src_ftp_attrs,
						  source_attrs);
    }
    
    if(rc != GLOBUS_SUCCESS)
    {
	goto destroy_collection_attrs;
    }

    if(dest_location_attrs == GLOBUS_NULL)
    {
	rc = globus_ftp_client_operationattr_init(&dest_ftp_attrs);
    }
    else
    {
	rc = globus_ftp_client_operationattr_copy(&dest_ftp_attrs,
						  dest_location_attrs);
    }
    
    if(rc != GLOBUS_SUCCESS)
    {
	goto destroy_src_attrs;
    }

    rc = globus_l_replica_management_get_catalog_handle(handle,
							dest_collection_url,
							&collection_attrs);
    if(rc != GLOBUS_SUCCESS)
    {
	goto destroy_dest_attrs;
    }

    rc = globus_l_replica_management_file_move(handle,
					       source_url,
					       &src_ftp_attrs,
					       dest_collection_url,
					       &collection_attrs,
					       dest_location_name,
					       &dest_ftp_attrs,
					       dest_filename,
					       overwrite_behavior,
					       preregister,
					       preregister_arg,
					       GLOBUS_FALSE);
 destroy_dest_attrs:
    globus_ftp_client_operationattr_destroy(&dest_ftp_attrs);
 destroy_src_attrs:
    globus_ftp_client_operationattr_destroy(&src_ftp_attrs);
 destroy_collection_attrs:
    globus_replica_catalog_collection_handleattr_destroy(&collection_attrs);
 return_result:
    return rc;
}

globus_result_t
globus_replica_management_file_copy(
    globus_replica_management_handle_t *              handle,
    char *                                            collection_url,
    globus_replica_catalog_collection_handleattr_t *  collection_attrs,
    char *                                            source_location_name,
    globus_ftp_client_operationattr_t *               source_location_attrs,
    char *                                            dest_location_name,
    globus_ftp_client_operationattr_t *               dest_location_attrs,
    char *                                            filename,
    globus_replica_management_overwrite_behavior_t    overwrite_behavior,
    globus_replica_management_preregister_function_t  preregister,
    void *                                            preregister_arg)
{
    globus_result_t                                     rc;
    globus_replica_catalog_collection_handleattr_t      coll_attrs;
    globus_ftp_client_operationattr_t                   src_ftp_attrs;
    globus_ftp_client_operationattr_t                   dest_ftp_attrs;
    char *                                              src_url;
    static char * myname = "globus_replica_management_file_copy";

    if(handle == GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "NULL handle",
	    myname);
	goto return_result;
    }

    if(handle->replica_handle != GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "Busy handle",
	    myname);
	goto return_result;
    }

    if(collection_url == GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "NULL URL",
	    myname);
	goto return_result;
    }

    if(source_location_name == GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "NULL location name",
	    myname);
	goto return_result;
    }
    
    if(dest_location_name == GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "NULL location name",
	    myname);
	goto return_result;
    }

    if(filename == GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "NULL file name",
	    myname);
	goto return_result;
    }

    if(overwrite_behavior != GLOBUS_REPLICA_MANAGEMENT_OVERWRITE_DISALLOW &&
       overwrite_behavior != GLOBUS_REPLICA_MANAGEMENT_OVERWRITE_SAFE     &&
       overwrite_behavior != GLOBUS_REPLICA_MANAGEMENT_OVERWRITE_UNSAFE)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "invalid overwrite behavior",
	    myname);
	goto return_result;
    }
    
    if(collection_attrs == GLOBUS_NULL)
    {
	rc = globus_replica_catalog_collection_handleattr_init(
	    &coll_attrs);
    }
    else
    {
	rc = globus_replica_catalog_collection_handleattr_copy(
	    &coll_attrs,
	    collection_attrs);
    }
    
    if(rc != GLOBUS_SUCCESS)
    {
	goto return_result;
    }

    if(source_location_attrs == GLOBUS_NULL)
    {
	rc = globus_ftp_client_operationattr_init(&src_ftp_attrs);
    }
    else
    {
	rc = globus_ftp_client_operationattr_copy(&src_ftp_attrs,
						  source_location_attrs);
    }
    
    if(rc != GLOBUS_SUCCESS)
    {
	goto destroy_collection_attrs;
    }

    if(dest_location_attrs == GLOBUS_NULL)
    {
	rc = globus_ftp_client_operationattr_init(&dest_ftp_attrs);
    }
    else
    {
	rc = globus_ftp_client_operationattr_copy(&dest_ftp_attrs,
						  dest_location_attrs);
    }
    
    if(rc != GLOBUS_SUCCESS)
    {
	goto destroy_src_attrs;
    }

    rc = globus_l_replica_management_get_catalog_handle(handle,
							collection_url,
							&coll_attrs);
    if(rc != GLOBUS_SUCCESS)
    {
	goto destroy_dest_attrs;
    }

    rc = globus_i_replica_catalog_location_get_url(
	handle->replica_handle,
	source_location_name,
	filename,
	&src_url);

    if(rc != GLOBUS_SUCCESS)
    {
	globus_l_replica_management_destroy_catalog_handle(handle);
	goto destroy_dest_attrs;
    }

    rc = globus_l_replica_management_file_move(handle,
					       src_url,
					       &src_ftp_attrs,
					       collection_url,
					       &coll_attrs,
					       dest_location_name,
					       &dest_ftp_attrs,
					       filename,
					       overwrite_behavior,
					       preregister,
					       preregister_arg,
					       GLOBUS_TRUE);
 destroy_src_url:
    globus_libc_free(src_url);
 destroy_dest_attrs:
    globus_ftp_client_operationattr_destroy(&dest_ftp_attrs);
 destroy_src_attrs:
    globus_ftp_client_operationattr_destroy(&src_ftp_attrs);
 destroy_collection_attrs:
    globus_replica_catalog_collection_handleattr_destroy(&coll_attrs);
 return_result:
    return rc;
}

globus_result_t
globus_replica_management_file_delete(
    globus_replica_management_handle_t *              handle,
    char *                                            collection_url,
    globus_replica_catalog_collection_handleattr_t *  collection_attrs,
    char *                                            location_name,
    globus_ftp_client_operationattr_t *               location_attrs,
    char *                                            filename,
    globus_bool_t                                     delete_file)
{
    globus_result_t                                     rc;
    globus_replica_catalog_collection_handleattr_t      coll_attrs;
    globus_ftp_client_operationattr_t                   ftp_attrs;
    globus_i_replica_management_monitor_t               monitor;
    char *                                              url = GLOBUS_NULL;
    char *                                              filenames[2];
    char *                                              rollback_contact;
    static char * myname = "globus_replica_management_file_delete";

    if(handle == GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "NULL handle",
	    myname);
	goto return_result;
    }

    if(handle->replica_handle != GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "Busy handle",
	    myname);
	goto return_result;
    }

    if(collection_url == GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "NULL URL",
	    myname);
	goto return_result;
    }

    if(location_name == GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "NULL location name",
	    myname);
	goto return_result;
    }
    
    if(filename == GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "NULL file name",
	    myname);
	goto return_result;
    }
    
    if(collection_attrs == GLOBUS_NULL)
    {
	rc = globus_replica_catalog_collection_handleattr_init(
	    &coll_attrs);
    }
    else
    {
	rc = globus_replica_catalog_collection_handleattr_copy(
	    &coll_attrs,
	    collection_attrs);
    }
    
    if(rc != GLOBUS_SUCCESS)
    {
	goto return_result;
    }

    if(location_attrs == GLOBUS_NULL)
    {
	rc = globus_ftp_client_operationattr_init(&ftp_attrs);
    }
    else
    {
	rc = globus_ftp_client_operationattr_copy(&ftp_attrs,
						  location_attrs);
    }
    
    if(rc != GLOBUS_SUCCESS)
    {
	goto destroy_collection_attrs;
    }

    rc = globus_l_replica_management_get_catalog_handle(handle,
							collection_url,
							&coll_attrs);
    if(rc != GLOBUS_SUCCESS)
    {
	goto  destroy_ftp_attrs;
    }
    
    rc = globus_i_replica_catalog_rollback_info_construct_contact(
	handle->replica_handle,
	location_name,
	filename,
	&rollback_contact);
    
    if(rc != GLOBUS_SUCCESS)
    {
	globus_l_replica_management_destroy_catalog_handle(handle);
	goto destroy_ftp_attrs;
    }
    
    if(delete_file == GLOBUS_TRUE)
    {
	rc = globus_i_replica_catalog_rollback_info_create(
	    handle->replica_handle,
	    location_name,
	    filename,
	    GLOBUS_I_REPLICA_MANAGEMENT_DELETE,
	    GLOBUS_I_REPLICA_MANAGEMENT_PHYSICAL_DELETE_START,
	    GLOBUS_NULL,
	    GLOBUS_NULL);
    }
    else
    {
	rc = globus_i_replica_catalog_rollback_info_create(
	    handle->replica_handle,
	    location_name,
	    filename,
	    GLOBUS_I_REPLICA_MANAGEMENT_DELETE,
	    GLOBUS_I_REPLICA_MANAGEMENT_LOGICAL_DELETE_START,
	    GLOBUS_NULL,
	    GLOBUS_NULL);
    }


    if(rc != GLOBUS_SUCCESS)
    {
	globus_l_replica_management_destroy_catalog_handle(handle);
	goto destroy_contact;
    }
    
    filenames[0] = filename;
    filenames[1] = GLOBUS_NULL;
    
    rc = GLOBUS_REPLICA_CATALOG_LOCATION_DELETE_FILENAMES(
	handle->replica_handle,
	location_name,
	filenames);

    if(rc != GLOBUS_SUCCESS)
    {
	/* would like to know if this error is due to the
	   absense of this attribute. If so we should not
	   return an error.
	*/
	
	rc = GLOBUS_REPLICA_MANAGEMENT_ROLLBACK_ERROR(
	    rollback_contact,
	    globus_error_get(rc));
	globus_l_replica_management_destroy_catalog_handle(handle);
	goto destroy_contact;
    }

    if(delete_file == GLOBUS_TRUE)
    {
	rc = GLOBUS_I_REPLICA_CATALOG_ROLLBACK_INFO_UPDATE(
	    handle->replica_handle,
	    location_name,
	    filename,
	    GLOBUS_I_REPLICA_MANAGEMENT_NEED_PHYSICAL_DELETE,
	    GLOBUS_NULL);

	if(rc != GLOBUS_SUCCESS)
	{
	    rc = GLOBUS_REPLICA_MANAGEMENT_ROLLBACK_ERROR(
		rollback_contact,
		globus_error_get(rc));
	    globus_l_replica_management_destroy_catalog_handle(handle);
	    goto destroy_contact;
	}
	
	rc = globus_i_replica_management_monitor_init(&monitor);
	
	if(rc != GLOBUS_SUCCESS)
	{
	    rc = GLOBUS_REPLICA_MANAGEMENT_ROLLBACK_ERROR(
		rollback_contact,
		globus_error_get(rc));

	    globus_l_replica_management_release_catalog_handle(handle);
	    goto destroy_contact;
	}

	rc = globus_i_replica_catalog_location_get_url(
	    handle->replica_handle,
	    location_name,
	    filename,
	    &url);

	if(rc != GLOBUS_SUCCESS)
	{
	    rc = GLOBUS_REPLICA_MANAGEMENT_ROLLBACK_ERROR(
		rollback_contact,
		globus_error_get(rc));
	    globus_l_replica_management_destroy_catalog_handle(handle);
	    goto destroy_monitor;
	}
    
	rc = globus_ftp_client_delete(&handle->ftp_handle,
				      url,
				      &ftp_attrs,
				      globus_l_replica_management_ftp_cmd_cb,
				      (void *) &monitor);
	if(rc != GLOBUS_SUCCESS)
	{
	    rc = GLOBUS_REPLICA_MANAGEMENT_ROLLBACK_ERROR(
		rollback_contact,
		globus_error_get(rc));

	    globus_l_replica_management_release_catalog_handle(handle);
	    goto destroy_url;
	}
	
	globus_mutex_lock(&monitor.mutex);
	{
	    while(monitor.done == GLOBUS_FALSE)
	    {
		globus_cond_wait(&monitor.cond, &monitor.mutex);
	    }
	}
	globus_mutex_unlock(&monitor.mutex);

	if(monitor.result != GLOBUS_SUCCESS)
	{
	    /* would like to know if this error is due to the
	       absense of the file. If so we should not
	       return an error.
	    */

	    rc = monitor.result;
	    rc = GLOBUS_REPLICA_MANAGEMENT_ROLLBACK_ERROR(
		rollback_contact,
		globus_error_get(rc));
	    goto destroy_url;
	}
    }

    rc = GLOBUS_I_REPLICA_CATALOG_ROLLBACK_INFO_DELETE(
	handle->replica_handle,
	location_name,
	filename);

    if(rc != GLOBUS_SUCCESS)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ROLLBACK_ERROR(
	    rollback_contact,
	    globus_error_get(rc));
	
	globus_l_replica_management_destroy_catalog_handle(handle);
	goto destroy_monitor;
    }
    
    rc = globus_l_replica_management_release_catalog_handle(handle);
    
 destroy_url:
    if(url != GLOBUS_NULL)
    {
	globus_libc_free(url);
    }
 destroy_monitor:
    if(delete_file == GLOBUS_TRUE)
    {
        globus_i_replica_management_monitor_destroy(&monitor);
    }
 destroy_contact:
    globus_libc_free(rollback_contact);
 destroy_ftp_attrs:
    globus_ftp_client_operationattr_destroy(&ftp_attrs);
 destroy_collection_attrs:
    globus_replica_catalog_collection_handleattr_destroy(&coll_attrs);
 return_result:
    return rc;
}

globus_result_t
globus_replica_management_restart(
    globus_replica_management_handle_t *              handle,
    char *                                            rollback_info,
    globus_replica_catalog_collection_handleattr_t *  collection_attrs,
    globus_ftp_client_operationattr_t *               source_ftp_attrs,
    globus_ftp_client_operationattr_t *               destination_ftp_attrs,
    globus_replica_management_preregister_function_t  preregister,
    void *                                            preregister_arg)
{
    globus_result_t                                     rc;
    globus_replica_catalog_collection_handleattr_t      coll_attrs;
    globus_ftp_client_operationattr_t                   src_ftp_attrs;
    globus_ftp_client_operationattr_t                   dest_ftp_attrs;
    globus_i_replica_management_monitor_t               monitor;
    globus_ftp_client_restart_marker_t     	        marker;
    globus_l_replica_management_restart_state_t         state;
    globus_bool_t                                       need_update = GLOBUS_FALSE;
    char *                                              url;
    char *                                              tmp_url = GLOBUS_NULL;
    char *                                              location_name;
    char *                                              filename;
    char *                                              collection_url;
    char *                                              operation;
    char *                                              catalog_state;
    char *                                              session_id;
    char *                                              source_url;
    char *                                              ranges;
    char *                                              filenames[2];
    globus_gass_copy_attr_t				source_gass_copy_attrs;
    globus_gass_copy_attr_t				dest_gass_copy_attrs;
    static char * myname = "globus_replica_management_restart";

    if(handle == GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "NULL handle",
	    myname);
	goto return_result;
    }

    if(handle->replica_handle != GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "Busy handle",
	    myname);
	goto return_result;
    }

    if(rollback_info == GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "NULL rollback information contact string",
	    myname);
	goto return_result;
    }

    if(collection_attrs == GLOBUS_NULL)
    {
	rc = globus_replica_catalog_collection_handleattr_init(&coll_attrs);
    }
    else
    {
	rc = globus_replica_catalog_collection_handleattr_copy(
	    &coll_attrs,
	    collection_attrs);
    }
    
    if(rc != GLOBUS_SUCCESS)
    {
	goto return_result;
    }


    if(source_ftp_attrs == GLOBUS_NULL)
    {
	rc = globus_ftp_client_operationattr_init(&src_ftp_attrs);
    }
    else
    {
	rc = globus_ftp_client_operationattr_copy(&src_ftp_attrs,
						  source_ftp_attrs);
    }
    
    if(rc != GLOBUS_SUCCESS)
    {
	goto destroy_collection_attrs;
    }

    if(destination_ftp_attrs == GLOBUS_NULL)
    {
	rc = globus_ftp_client_operationattr_init(&dest_ftp_attrs);
    }
    else
    {
	rc = globus_ftp_client_operationattr_copy(&dest_ftp_attrs,
						  destination_ftp_attrs);
    }
    
    if(rc != GLOBUS_SUCCESS)
    {
	goto destroy_src_attrs;
    }

    rc = globus_i_replica_catalog_rollback_info_parse_contact(rollback_info,
							      &collection_url,
							      &location_name,
							      &filename);
    if(rc != GLOBUS_SUCCESS)
    {
	goto destroy_dest_attrs;
    }

    rc = globus_l_replica_management_get_catalog_handle(handle,
							collection_url,
							&coll_attrs);
    if(rc != GLOBUS_SUCCESS)
    {
	goto destroy_contact_info;
    }

    rc = globus_i_replica_catalog_rollback_info_get(handle->replica_handle,
						    location_name,
						    filename,
						    &operation,
						    &catalog_state,
						    &session_id,
						    &source_url,
						    &ranges);
    if(rc != GLOBUS_SUCCESS)
    {
	globus_l_replica_management_destroy_catalog_handle(handle);
	goto destroy_contact_info;
    }


    rc = globus_i_replica_management_monitor_init(&monitor);
    
    if(rc != GLOBUS_SUCCESS)
    {
	globus_l_replica_management_release_catalog_handle(handle);
	goto destroy_rollback_info;
    }
    
    rc = globus_i_replica_management_set_plugin_data(handle,
						     location_name,
						     filename);
    
    if(rc != GLOBUS_SUCCESS)
    {
	globus_l_replica_management_release_catalog_handle(handle);
	goto destroy_monitor;
    }

    rc = globus_i_replica_catalog_location_get_url(
	handle->replica_handle,
	location_name,
	filename,
	&url);
    
    if(rc != GLOBUS_SUCCESS)
    {
	globus_l_replica_management_destroy_catalog_handle(handle);
	goto destroy_monitor;
    }
    
    if(session_id != GLOBUS_NULL)
    {
	tmp_url = globus_libc_malloc(strlen(url) +
				     strlen(session_id) + 5);
	if(tmp_url == GLOBUS_NULL)
	{
	    rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
		"Could not allocate internal data structure",
		myname);
	    globus_l_replica_management_release_catalog_handle(handle);
	    goto destroy_url;
	}
	
	strcpy(tmp_url, url);
	strcat(tmp_url, ".");
	strcat(tmp_url, session_id);
    }
    
    state = globus_l_replica_management_get_state(catalog_state);

 restart:    
    switch(state)
    {
    case GLOBUS_L_REPLICA_MANAGEMENT_PHYSICAL_DELETE_START:
    case GLOBUS_L_REPLICA_MANAGEMENT_LOGICAL_DELETE_START:
	filenames[0] = filename;
	filenames[1] = GLOBUS_NULL;
	
	rc = globus_replica_catalog_location_delete_filenames(
	    handle->replica_handle,
	    location_name,
	    filenames);
	
	if(rc != GLOBUS_SUCCESS)
	{
	    /* would like to know what error occured */
	}
	
	if(state == GLOBUS_L_REPLICA_MANAGEMENT_PHYSICAL_DELETE_START)
	{
	    need_update = GLOBUS_TRUE;
	    state = GLOBUS_L_REPLICA_MANAGEMENT_NEED_PHYSICAL_DELETE;
	    goto restart;
	}

	break;
    case GLOBUS_L_REPLICA_MANAGEMENT_NEED_PHYSICAL_DELETE:

	if(need_update == GLOBUS_TRUE)
	{
	    rc = GLOBUS_I_REPLICA_CATALOG_ROLLBACK_INFO_UPDATE(
		handle->replica_handle,
		location_name,
		filename,
		GLOBUS_I_REPLICA_MANAGEMENT_NEED_PHYSICAL_DELETE,
		GLOBUS_NULL);
	
	    if(rc != GLOBUS_SUCCESS)
	    {
		globus_l_replica_management_release_catalog_handle(handle);
		goto destroy_tmp_url;
	    }
	}
	
	rc = globus_ftp_client_delete(
	    &handle->ftp_handle,
	    url,
	    &dest_ftp_attrs,
	    globus_l_replica_management_ftp_cmd_cb,
	    (void *) &monitor);
	    
	if(rc != GLOBUS_SUCCESS)
	{
	    globus_l_replica_management_release_catalog_handle(handle);
	    goto destroy_tmp_url;
	}
	
	globus_mutex_lock(&monitor.mutex);
	{
	    while(monitor.done == GLOBUS_FALSE)
	    {
		globus_cond_wait(&monitor.cond, &monitor.mutex);
	    }
	}
	globus_mutex_unlock(&monitor.mutex);
	
	if(monitor.result != GLOBUS_SUCCESS)
	{
	    /* need to know what error occured */
	}

	break;
	
    case GLOBUS_L_REPLICA_MANAGEMENT_UNSAFE_START:
    case GLOBUS_L_REPLICA_MANAGEMENT_SAFE_START:

	/* restart the transfer */

	/* if ranges is NULL or ranges is
	   not NULL and ranges is not done */
	    
	if(ranges == GLOBUS_NULL ||
	   strcmp(ranges, "done"))
	{
	    if(ranges != GLOBUS_NULL)
	    {
		rc = globus_ftp_client_restart_marker_from_string(&marker,
								  ranges);
		if(rc != GLOBUS_SUCCESS)
		{
		    globus_l_replica_management_release_catalog_handle(
			handle);
		    goto destroy_tmp_url;
		}
	    
		/* XXX: if gass_copy takes restart markers the
		 * following will be obsoleted */

		rc = globus_ftp_client_third_party_transfer(
		    &handle->ftp_handle,
		    source_url,
		    &src_ftp_attrs,
		    tmp_url != GLOBUS_NULL ? tmp_url : url, 
		    &dest_ftp_attrs,
		    &marker,
		    globus_l_replica_management_ftp_cmd_cb,
		    (void *) &monitor);

		if(rc != GLOBUS_SUCCESS)
		{
		    globus_l_replica_management_destroy_catalog_handle(handle);
		    goto destroy_marker;
		}
		
		globus_mutex_lock(&monitor.mutex);
		{
		    while(monitor.done == GLOBUS_FALSE)
		    {
			globus_cond_wait(&monitor.cond, &monitor.mutex);
		    }
		}
		globus_mutex_unlock(&monitor.mutex);
		
		if(monitor.result != GLOBUS_SUCCESS)
		{
		    rc = monitor.result;
		    globus_l_replica_management_release_catalog_handle(handle);
		    goto destroy_marker;
		}

	    }
	    else
	    {

		/* XXX: for now, gass_copy only handles the case where
		 * there are no existing ranges to restart with */ 

		rc = globus_gass_copy_attr_init(&source_gass_copy_attrs);
		if (rc != GLOBUS_SUCCESS)
		{
		    globus_l_replica_management_release_catalog_handle(handle);
		    goto destroy_tmp_url;
		}

		rc = globus_gass_copy_attr_set_ftp(&source_gass_copy_attrs,
						&src_ftp_attrs);
		if (rc != GLOBUS_SUCCESS)
		{
		    globus_l_replica_management_release_catalog_handle(handle);
		    goto destroy_tmp_url;
		}

		rc = globus_gass_copy_attr_init(&dest_gass_copy_attrs);
		if (rc != GLOBUS_SUCCESS)
		{
		    globus_l_replica_management_release_catalog_handle(handle);
		    goto destroy_tmp_url;
		}

		rc = globus_gass_copy_attr_set_ftp(&dest_gass_copy_attrs,
						&dest_ftp_attrs);
		if (rc != GLOBUS_SUCCESS)
		{
		    globus_l_replica_management_release_catalog_handle(handle);
		    goto destroy_tmp_url;
		}

		rc = globus_gass_copy_register_url_to_url(
			&handle->gass_copy_handle,
			source_url,
			&source_gass_copy_attrs,
			tmp_url != GLOBUS_NULL ? tmp_url : url, 
			&dest_gass_copy_attrs,
			globus_l_gass_copy_callback,
			(void *) &monitor);
		if (rc != GLOBUS_SUCCESS)
		{
		    globus_l_replica_management_destroy_catalog_handle(handle);
		    goto destroy_marker;
		}

		globus_mutex_lock(&monitor.mutex);
		{
		    while(monitor.done == GLOBUS_FALSE)
		    {
			globus_cond_wait(&monitor.cond, &monitor.mutex);
		    }
		}
		globus_mutex_unlock(&monitor.mutex);

		if(monitor.result != GLOBUS_SUCCESS)
		{
		    rc = monitor.result;
		    globus_l_replica_management_release_catalog_handle(handle);
		    goto destroy_marker;
		}
	    }

	}

	if(session_id != GLOBUS_NULL)
	{
	    state = GLOBUS_L_REPLICA_MANAGEMENT_NEED_FILE_MOVE;
	}
	else
	{
	    state = GLOBUS_L_REPLICA_MANAGEMENT_NEED_PREREGISTER;    
	}

	need_update = GLOBUS_TRUE;
	
	goto restart;
	
    case GLOBUS_L_REPLICA_MANAGEMENT_NEED_FILE_MOVE:

	if(need_update == GLOBUS_TRUE)
	{
	    rc = GLOBUS_I_REPLICA_CATALOG_ROLLBACK_INFO_UPDATE(
		handle->replica_handle,
		location_name,
		filename,
		GLOBUS_I_REPLICA_MANAGEMENT_NEED_FILE_MOVE,
		GLOBUS_NULL);
		
	    if(rc != GLOBUS_SUCCESS)
	    {
		globus_l_replica_management_destroy_catalog_handle(handle);
		goto destroy_marker;
	    }
	}
	    
	globus_i_replica_management_monitor_reset(&monitor);
	
	rc = globus_ftp_client_move(
	    &handle->ftp_handle,
	    tmp_url,
	    url,
	    &dest_ftp_attrs,
	    globus_l_replica_management_ftp_cmd_cb,
	    (void *) &monitor);
	
	if(rc != GLOBUS_SUCCESS)
	{
	    globus_l_replica_management_release_catalog_handle(handle);
	    goto destroy_marker;
	}
	
	globus_mutex_lock(&monitor.mutex);
	{
	    while(monitor.done == GLOBUS_FALSE)
	    {
		globus_cond_wait(&monitor.cond, &monitor.mutex);
	    }
	}
	globus_mutex_unlock(&monitor.mutex);
	
	if(monitor.result != GLOBUS_SUCCESS)
	{
	    /* Need to know exactly what error occured */
	}

	state = GLOBUS_L_REPLICA_MANAGEMENT_NEED_PREREGISTER;

	need_update = GLOBUS_TRUE;
	
	goto restart;
	
    case GLOBUS_L_REPLICA_MANAGEMENT_NEED_PREREGISTER:

	if(need_update == GLOBUS_TRUE)
	{
	    rc = GLOBUS_I_REPLICA_CATALOG_ROLLBACK_INFO_UPDATE(
		handle->replica_handle,
		location_name,
		filename,
		GLOBUS_I_REPLICA_MANAGEMENT_NEED_PREREGISTER,
		GLOBUS_NULL);
		
	    if(rc != GLOBUS_SUCCESS)
	    {
		globus_l_replica_management_destroy_catalog_handle(
		    handle);
		goto destroy_marker;
	    }
	}
	
	if(preregister != GLOBUS_NULL)
	{
	    rc = preregister(preregister_arg, url);

	    if(rc != GLOBUS_SUCCESS)
	    {
		globus_l_replica_management_release_catalog_handle(handle);
		goto destroy_marker;
	    }
	}

	if(!strcmp(operation, GLOBUS_I_REPLICA_MANAGEMENT_PUBLISH))
	{
	    state = GLOBUS_L_REPLICA_MANAGEMENT_NEED_COLLECTION_ADD;
	}
	else
	{
	    state = GLOBUS_L_REPLICA_MANAGEMENT_NEED_LOCATION_ADD;
	}
	
	need_update = GLOBUS_TRUE;
	
	goto restart;

    case GLOBUS_L_REPLICA_MANAGEMENT_NEED_COLLECTION_ADD:
	if(need_update == GLOBUS_TRUE)
	{
	    rc = GLOBUS_I_REPLICA_CATALOG_ROLLBACK_INFO_UPDATE(
		handle->replica_handle,
		location_name,
		filename,
		GLOBUS_I_REPLICA_MANAGEMENT_NEED_COLLECTION_ADD,
		GLOBUS_NULL);
	
	    if(rc != GLOBUS_SUCCESS)
	    {
		globus_l_replica_management_destroy_catalog_handle(handle);
		goto destroy_marker;
	    }
	}

	filenames[0] = filename;
	filenames[1] = GLOBUS_NULL;
	
	rc = globus_replica_catalog_collection_add_filenames(
	    handle->replica_handle,
	    filenames,
	    GLOBUS_TRUE);
	
	if(rc != GLOBUS_SUCCESS)
	{
	    globus_l_replica_management_destroy_catalog_handle(handle);
	    goto destroy_marker;
	}

	state = GLOBUS_L_REPLICA_MANAGEMENT_NEED_LOCATION_ADD;

	need_update = GLOBUS_TRUE;
	
	goto restart;
	
    case GLOBUS_L_REPLICA_MANAGEMENT_NEED_LOCATION_ADD:
	if(need_update == GLOBUS_TRUE)
	{
	    rc = GLOBUS_I_REPLICA_CATALOG_ROLLBACK_INFO_UPDATE(
		handle->replica_handle,
		location_name,
		filename,
		GLOBUS_I_REPLICA_MANAGEMENT_NEED_LOCATION_ADD,
		GLOBUS_NULL);

	    if(rc != GLOBUS_SUCCESS)
	    {
		globus_l_replica_management_destroy_catalog_handle(handle);
		goto destroy_marker;
	    }
	}

	filenames[0] = filename;
	filenames[1] = GLOBUS_NULL;
	
	rc = globus_replica_catalog_location_add_filenames(
	    handle->replica_handle,
	    location_name,
	    filenames,
	    GLOBUS_TRUE);
	
	if(rc != GLOBUS_SUCCESS)
	{
	    globus_l_replica_management_destroy_catalog_handle(handle);
	    goto destroy_marker;
	}

	break;
    }
    
    rc = GLOBUS_I_REPLICA_CATALOG_ROLLBACK_INFO_DELETE(handle->replica_handle,
						       location_name,
						       filename);
    if(rc != GLOBUS_SUCCESS)
    {
	globus_l_replica_management_destroy_catalog_handle(handle);
	goto destroy_marker;
    }

    rc = globus_l_replica_management_release_catalog_handle(handle);
    
 destroy_marker:
    if((ranges != GLOBUS_NULL) && !strcmp(ranges, "done"))
    {
	globus_ftp_client_restart_marker_destroy(&marker);
    }
 destroy_tmp_url:
    if(tmp_url != GLOBUS_NULL)
    {
	globus_libc_free(tmp_url);
    } 
 destroy_url:
    globus_libc_free(url);
 destroy_monitor:
    globus_i_replica_management_monitor_destroy(&monitor);
 destroy_rollback_info:
    globus_libc_free(operation);
    globus_libc_free(catalog_state);
    if(session_id != GLOBUS_NULL)
    {
	globus_libc_free(session_id);
    }

    if(source_url != GLOBUS_NULL)
    {
	globus_libc_free(source_url);
    }

    if(ranges != GLOBUS_NULL)
    {
	globus_libc_free(ranges);
    }
 destroy_contact_info:
    globus_libc_free(collection_url);
    globus_libc_free(location_name);
    globus_libc_free(filename);
 destroy_dest_attrs:
    globus_ftp_client_operationattr_destroy(&dest_ftp_attrs);
 destroy_src_attrs:
    globus_ftp_client_operationattr_destroy(&src_ftp_attrs);
 destroy_collection_attrs:
    globus_replica_catalog_collection_handleattr_destroy(&coll_attrs);
 return_result:
    return rc;
}

globus_result_t
globus_replica_management_rollback(
    globus_replica_management_handle_t *              handle,
    char *                                            rollback_info,
    globus_replica_catalog_collection_handleattr_t *  collection_attrs,
    globus_ftp_client_operationattr_t *               ftp_attrs)
{
    globus_result_t                                     rc;
    globus_replica_catalog_collection_handleattr_t      coll_attrs;
    globus_ftp_client_operationattr_t                   my_ftp_attrs;
    globus_i_replica_management_monitor_t               monitor;
    globus_bool_t		                        deleted_tmp_file = GLOBUS_FALSE;
    globus_byte_t                                       buffer[GLOBUS_L_REPLICA_MANAGEMENT_BUFFER_SIZE];
    char *                                              url;
    char *                                              tmp_url = GLOBUS_NULL;
    char *                                              location_name;
    char *                                              filename;
    char *                                              collection_url;
    char *                                              operation;
    char *                                              state;
    char *                                              session_id;
    char *                                              source_url;
    char *                                              ranges;
    char *                                              filenames[2];
    static char * myname = "globus_replica_management_rollback";

    if(handle == GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "NULL handle",
	    myname);
	goto return_result;
    }

    if(handle->replica_handle != GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "Busy handle",
	    myname);
	goto return_result;
    }

    if(rollback_info == GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "NULL rollback information contact string",
	    myname);
	goto return_result;
    }

    if(collection_attrs == GLOBUS_NULL)
    {
	rc = globus_replica_catalog_collection_handleattr_init(&coll_attrs);
    }
    else
    {
	rc = globus_replica_catalog_collection_handleattr_copy(
	    &coll_attrs,
	    collection_attrs);
    }
    
    if(rc != GLOBUS_SUCCESS)
    {
	goto return_result;
    }


    if(ftp_attrs == GLOBUS_NULL)
    {
	rc = globus_ftp_client_operationattr_init(&my_ftp_attrs);
    }
    else
    {
	rc = globus_ftp_client_operationattr_copy(&my_ftp_attrs,
						  ftp_attrs);
    }
    
    if(rc != GLOBUS_SUCCESS)
    {
	goto destroy_collection_attrs;
    }

    rc = globus_i_replica_catalog_rollback_info_parse_contact(rollback_info,
							      &collection_url,
							      &location_name,
							      &filename);
    if(rc != GLOBUS_SUCCESS)
    {
	goto destroy_ftp_attrs;
    }

    rc = globus_l_replica_management_get_catalog_handle(handle,
							collection_url,
							&coll_attrs);
    if(rc != GLOBUS_SUCCESS)
    {
	goto destroy_contact_info;
    }

    rc = globus_i_replica_catalog_rollback_info_get(handle->replica_handle,
						    location_name,
						    filename,
						    &operation,
						    &state,
						    &session_id,
						    &source_url,
						    &ranges);
    if(rc != GLOBUS_SUCCESS)
    {
	globus_l_replica_management_destroy_catalog_handle(handle);
	goto destroy_contact_info;
    }

    if(!strcmp(state, GLOBUS_I_REPLICA_MANAGEMENT_PHYSICAL_DELETE_START) ||
       !strcmp(state, GLOBUS_I_REPLICA_MANAGEMENT_LOGICAL_DELETE_START))
    {
	/* try adding the filename to the location again */
	
	filenames[0] = filename;
	filenames[1] = GLOBUS_NULL;
	
	rc = globus_replica_catalog_location_add_filenames(
	    handle->replica_handle,
	    location_name,
	    filenames,
	    GLOBUS_TRUE);
	
	if(rc != GLOBUS_SUCCESS)
	{
	    globus_l_replica_management_destroy_catalog_handle(handle);
	    goto destroy_rollback_info;
	}
	
	rc = GLOBUS_I_REPLICA_CATALOG_ROLLBACK_INFO_DELETE(
	    handle->replica_handle,
	    location_name,
	    filename);

	globus_l_replica_management_release_catalog_handle(handle);
	goto destroy_rollback_info;

    }
    else if(!strcmp(state, GLOBUS_I_REPLICA_MANAGEMENT_NEED_PHYSICAL_DELETE))
    {
	/* Check if file is still around. If so add to location again */

	rc = globus_i_replica_management_monitor_init(&monitor);

	if(rc != GLOBUS_SUCCESS)
	{
	    globus_l_replica_management_release_catalog_handle(handle);
	    goto destroy_rollback_info;
	}

	rc = globus_i_replica_catalog_location_get_url(
	    handle->replica_handle,
	    location_name,
	    filename,
	    &url);
	
	if(rc != GLOBUS_SUCCESS)
	{
	    globus_l_replica_management_destroy_catalog_handle(handle);
	    goto destroy_monitor;
	}
    
        rc = globus_ftp_client_exists(&handle->ftp_handle,
                                      url,
                                      &my_ftp_attrs,
                                      globus_l_replica_management_ftp_exists_cb,
                                      (void *) &monitor);
	
	if(rc != GLOBUS_SUCCESS)
	{
	    globus_l_replica_management_release_catalog_handle(handle);
	    goto destroy_url;
	}

	globus_mutex_lock(&monitor.mutex);
	{
	    while(monitor.done == GLOBUS_FALSE)
	    {
		globus_cond_wait(&monitor.cond, &monitor.mutex);
	    }
	}
	globus_mutex_unlock(&monitor.mutex);
	
	if(monitor.result != GLOBUS_SUCCESS)
	{
	    rc = monitor.result;

	    /* Really want to know what kind of error occured */
	    
	    GLOBUS_I_REPLICA_CATALOG_ROLLBACK_INFO_DELETE(
		handle->replica_handle,
		location_name,
		filename);
	    
	    globus_l_replica_management_release_catalog_handle(handle);
	    goto destroy_url;
	}

	filenames[0] = filename;
	filenames[1] = GLOBUS_NULL;
    
	rc = globus_replica_catalog_location_add_filenames(
	    handle->replica_handle,
	    location_name,
	    filenames,
	    GLOBUS_TRUE);

	if(rc != GLOBUS_SUCCESS)
	{
	    globus_l_replica_management_destroy_catalog_handle(handle);
	    goto destroy_url;
	}

	rc = GLOBUS_I_REPLICA_CATALOG_ROLLBACK_INFO_DELETE(
	    handle->replica_handle,
	    location_name,
	    filename);

	globus_l_replica_management_release_catalog_handle(handle);
	goto destroy_url;
    }
    else 
    {
	/* Check if copy was started. If so clean up
	 * Check if catalog was modified. If so clean up
	 */
	
	rc = globus_i_replica_management_monitor_init(&monitor);
	
	if(rc != GLOBUS_SUCCESS)
	{
	    globus_l_replica_management_release_catalog_handle(handle);
	    goto destroy_rollback_info;
	}
	
	rc = globus_i_replica_catalog_location_get_url(
	    handle->replica_handle,
	    location_name,
	    filename,
	    &url);
	
	if(rc != GLOBUS_SUCCESS)
	{
	    globus_l_replica_management_destroy_catalog_handle(handle);
	    goto destroy_monitor;
	}


	if(session_id != GLOBUS_NULL)
	{
	    tmp_url = globus_libc_malloc(sizeof(char) *
					 (strlen(url) +
					  strlen(session_id) + 2));
	    if(tmp_url == GLOBUS_NULL)
	    {
		rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
		    "Could not allocate internal data structure",
		    myname);
		globus_l_replica_management_release_catalog_handle(handle);
		goto destroy_url;
	    }
	    
	    strcpy(tmp_url, url);
	    strcat(tmp_url, ".");
	    strcat(tmp_url, session_id);
	    
            rc = globus_ftp_client_exists(
                &handle->ftp_handle,
                tmp_url,
                &my_ftp_attrs,
                globus_l_replica_management_ftp_exists_cb,
                (void *) &monitor);

	    if(rc != GLOBUS_SUCCESS)
	    {
		globus_l_replica_management_release_catalog_handle(handle);
		goto destroy_url;
	    }
	    
	    globus_mutex_lock(&monitor.mutex);
	    {
		while(monitor.done == GLOBUS_FALSE)
		{
		    globus_cond_wait(&monitor.cond, &monitor.mutex);
		}
	    }
	    globus_mutex_unlock(&monitor.mutex);
	    
	    /* if file exists, remove file */
	    
	    if(monitor.result == GLOBUS_SUCCESS)
	    {
		globus_i_replica_management_monitor_reset(&monitor);
		
		rc = globus_ftp_client_delete(
		    &handle->ftp_handle,
		    tmp_url,
		    &my_ftp_attrs,
		    globus_l_replica_management_ftp_cmd_cb,
		    (void *) &monitor);
		
		if(rc != GLOBUS_SUCCESS)
		{
		    globus_l_replica_management_release_catalog_handle(handle);
		    goto destroy_tmp_url;
		}
		
		globus_mutex_lock(&monitor.mutex);
		{
		    while(monitor.done == GLOBUS_FALSE)
		    {
			globus_cond_wait(&monitor.cond, &monitor.mutex);
		    }
		}
		globus_mutex_unlock(&monitor.mutex);
		
		if(monitor.result != GLOBUS_SUCCESS)
		{
		    rc = monitor.result;
		    globus_l_replica_management_release_catalog_handle(handle);
		    goto destroy_tmp_url;
		}

		deleted_tmp_file = GLOBUS_TRUE;

		rc = GLOBUS_I_REPLICA_CATALOG_ROLLBACK_INFO_DELETE(
		    handle->replica_handle,
		    location_name,
		    filename);

		globus_l_replica_management_release_catalog_handle(handle);
		goto destroy_tmp_url;

	    }

	}

	/* If we didn't delete the temporary file and the state is not
	   SAFE_START and session_id is GLOBUS_NULL, try to delete the
	   file
	*/
	
	if(deleted_tmp_file == GLOBUS_FALSE &&
	   !(!strcmp(state, GLOBUS_I_REPLICA_MANAGEMENT_SAFE_START) &&
	    session_id != GLOBUS_NULL))
	{
	    globus_i_replica_management_monitor_reset(&monitor);
	    
            rc = globus_ftp_client_exists(
                &handle->ftp_handle,
                url,
                &my_ftp_attrs,
                globus_l_replica_management_ftp_exists_cb,
                (void *) &monitor);
	    
	    if(rc != GLOBUS_SUCCESS)
	    {
		globus_l_replica_management_release_catalog_handle(handle);
		goto destroy_tmp_url;
	    }
	    
	    globus_mutex_lock(&monitor.mutex);
	    {
		while(monitor.done == GLOBUS_FALSE)
		{
		    globus_cond_wait(&monitor.cond, &monitor.mutex);
		}
	    }
	    globus_mutex_unlock(&monitor.mutex);
	    
	    /* if file exists, remove file */
	    
	    if(monitor.result == GLOBUS_SUCCESS)
	    {
		globus_i_replica_management_monitor_reset(&monitor);
		
		rc = globus_ftp_client_delete(
		    &handle->ftp_handle,
		    url,
		    &my_ftp_attrs,
		    globus_l_replica_management_ftp_cmd_cb,
		    (void *) &monitor);
		
		if(rc != GLOBUS_SUCCESS)
		{
		    globus_l_replica_management_release_catalog_handle(handle);
		    goto destroy_tmp_url;
		}
		
		globus_mutex_lock(&monitor.mutex);
		{
		    while(monitor.done == GLOBUS_FALSE)
		    {
			globus_cond_wait(&monitor.cond, &monitor.mutex);
		    }
		}
		globus_mutex_unlock(&monitor.mutex);
		
		if(monitor.result != GLOBUS_SUCCESS)
		{
		    rc = monitor.result;
		    globus_l_replica_management_release_catalog_handle(handle);
		    goto destroy_tmp_url;
		}

		/* we have deleted the file, now try to clean up the catalog
		 * This should maybe be done first?
		 */

		filenames[0] = filename;
		filenames[1] = GLOBUS_NULL;
		
		rc = globus_replica_catalog_location_delete_filenames(
		    handle->replica_handle,
		    location_name,
		    filenames);

		if(rc == GLOBUS_SUCCESS)
		{
		    /* Ignore? Need to discuss this */
		}

/*		rc = globus_replica_catalog_collection_delete_filenames(
		    handle->replica_handle,
		    filenames);
		
		if(rc == GLOBUS_SUCCESS)
		{
		}
*/
		rc = GLOBUS_I_REPLICA_CATALOG_ROLLBACK_INFO_DELETE(
		    handle->replica_handle,
		    location_name,
		    filename);

		globus_l_replica_management_release_catalog_handle(handle);
		goto destroy_tmp_url;
	    }
	}
    }
    
    
 destroy_tmp_url:
    if(tmp_url != GLOBUS_NULL)
    {
	globus_libc_free(tmp_url);
    } 
 destroy_url:
    globus_libc_free(url);
 destroy_monitor:
    globus_i_replica_management_monitor_destroy(&monitor);
 destroy_rollback_info:
    globus_libc_free(operation);
    globus_libc_free(state);
    if(session_id != GLOBUS_NULL)
    {
	globus_libc_free(session_id);
    }

    if(source_url != GLOBUS_NULL)
    {
	globus_libc_free(source_url);
    }

    if(ranges != GLOBUS_NULL)
    {
	globus_libc_free(ranges);
    }
 destroy_contact_info:
    globus_libc_free(collection_url);
    globus_libc_free(location_name);
    globus_libc_free(filename);
 destroy_ftp_attrs:
    globus_ftp_client_operationattr_destroy(&my_ftp_attrs);
 destroy_collection_attrs:
    globus_replica_catalog_collection_handleattr_destroy(&coll_attrs);
 return_result:
    return rc;
}


static globus_result_t
globus_l_replica_management_add_cache_entry(
    globus_replica_management_handle_t *                handle,
    char *                                              collection_url,
    globus_replica_catalog_collection_handleattr_t *    collection_attrs)
{
    globus_result_t                             rc;
    globus_l_replica_management_cache_entry_t * entry;
    static char * myname = "globus_l_replica_management_add_cache_entry";
    
    entry = globus_libc_malloc(
	sizeof(globus_l_replica_management_cache_entry_t));

    if(entry == GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "Could not allocate internal data structure",
	    myname);
	goto return_result;
    }

    entry->handle = handle->replica_handle;

    entry->url = globus_libc_strdup(collection_url);

    if(entry->url == GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "Could not allocate internal data structure",
	    myname);
	goto destroy_entry;
    }

    rc = globus_replica_catalog_collection_handleattr_copy(&entry->attrs,
							   collection_attrs);
    if(rc != GLOBUS_SUCCESS)
    {
	goto destroy_url;
    }
    
    globus_list_insert(&handle->replica_cache,
		       (void *) entry);

    handle->cache_entry = handle->replica_cache;

    return GLOBUS_SUCCESS;

 destroy_url:
    globus_libc_free(entry->url);
 destroy_entry:
    globus_libc_free(entry);
 return_result:
    return rc;
}



static globus_result_t
globus_l_replica_management_get_catalog_handle(
    globus_replica_management_handle_t *                handle,
    char *                                              collection_url,
    globus_replica_catalog_collection_handleattr_t *    collection_attrs)
{
    globus_result_t                             rc;
    globus_l_replica_management_cache_entry_t   entry;
    globus_list_t *                             element;
    static char * myname = "globus_l_replica_management_get_catalog_handle";

    if(handle->attrs.caching == GLOBUS_TRUE)
    {
	entry.url = collection_url;
	rc = globus_replica_catalog_collection_handleattr_copy(
	    &entry.attrs,
	    collection_attrs);

	if(rc != GLOBUS_SUCCESS)
	{
	    goto return_result;
	}
	
	element = globus_list_search_pred(
	    handle->replica_cache,
	    globus_l_replica_management_catalog_handle_compare,
	    (void *) &entry);

	globus_replica_catalog_collection_handleattr_destroy(&entry.attrs);
	
	if(element != GLOBUS_NULL)
	{
	    handle->replica_handle =
		((globus_l_replica_management_cache_entry_t *)
		 globus_list_first(element))->handle;
	    handle->cache_entry = element;
	    return GLOBUS_SUCCESS;
	}
    }
	
    handle->replica_handle = globus_libc_malloc(
	sizeof(globus_replica_catalog_collection_handle_t));
    
    if(handle->replica_handle == GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "Could not allocate internal data structure",
	    myname);
	goto return_result;
    }
    
    rc = globus_replica_catalog_collection_open(handle->replica_handle,
						collection_attrs,
						collection_url);
    if(rc != GLOBUS_SUCCESS)
    {
	goto free_handle;
    }

    if(handle->attrs.caching == GLOBUS_TRUE)
    {
	rc = globus_l_replica_management_add_cache_entry(handle,
							 collection_url,
	                                                 collection_attrs);
	if(rc != GLOBUS_SUCCESS)
	{
	    goto close_handle;
	}
    }

    return GLOBUS_SUCCESS;

 close_handle:
    globus_replica_catalog_collection_close(handle->replica_handle);
 free_handle:
    globus_libc_free(handle->replica_handle);
    handle->replica_handle = GLOBUS_NULL;
 return_result:
    return rc;    
}


static globus_result_t
globus_l_replica_management_release_catalog_handle(
    globus_replica_management_handle_t *    handle)
{
    globus_result_t                             rc = GLOBUS_SUCCESS;
    static char * myname = "globus_l_replica_management_release_catalog_handle";
    if(handle->replica_handle != GLOBUS_NULL)
    {
	if(handle->attrs.caching == GLOBUS_FALSE)
	{
	    rc = globus_replica_catalog_collection_close(
		handle->replica_handle);
	    globus_libc_free(handle->replica_handle);
	    handle->replica_handle = GLOBUS_NULL;
	}
	else
	{
	    handle->replica_handle = GLOBUS_NULL;
	}
    }
    
    return rc;
}


static globus_result_t
globus_l_replica_management_destroy_catalog_handle(
    globus_replica_management_handle_t *    handle)
{
    globus_result_t                             rc;
    globus_l_replica_management_cache_entry_t * entry;
    static char * myname = "globus_l_replica_management_destroy_catalog_handle";
    if(handle->attrs.caching == GLOBUS_TRUE)
    {
	entry = (globus_l_replica_management_cache_entry_t *)
	    globus_list_remove(&handle->replica_cache,
			       handle->cache_entry);

	rc = globus_replica_catalog_collection_close(entry->handle);
	globus_replica_catalog_collection_handleattr_destroy(&entry->attrs);
	globus_libc_free(entry->url);
	globus_libc_free(entry);
	handle->cache_entry = GLOBUS_NULL;
    }
    else
    {
	rc = globus_replica_catalog_collection_close(handle->replica_handle);
	globus_libc_free(handle->replica_handle);
    }

    handle->replica_handle = GLOBUS_NULL;
    return rc;
}


static globus_result_t
globus_l_replica_management_flush_catalog_cache(
    globus_replica_management_handle_t *    handle)
{
    globus_result_t                             rc = GLOBUS_SUCCESS;
    static char * myname = "globus_l_replica_management_flush_catalog_cache";


    
    while(!globus_list_empty(handle->replica_cache))
    {
	handle->cache_entry = handle->replica_cache;
	rc = globus_l_replica_management_destroy_catalog_handle(handle);
	if(rc != GLOBUS_SUCCESS)
	{
	    break;
	}
    }

    return rc;
}


static int
globus_l_replica_management_catalog_handle_compare(
    void *                              entry1,
    void *                              entry2)
{
    globus_l_replica_management_cache_entry_t * entry;
    globus_l_replica_management_cache_entry_t * desired_entry;

    entry = (globus_l_replica_management_cache_entry_t *) entry1;
    desired_entry = (globus_l_replica_management_cache_entry_t *) entry2;

    if(strcmp(entry->url,desired_entry->url) != 0)
    {
	return 1;
    }

    if(globus_i_replica_catalog_collection_handleattr_compare(
	&entry->attrs,
	&desired_entry->attrs) != 0)
    {
	return 1;
    }

    return 0;
}

static void
globus_l_replica_management_ftp_exists_cb(
    void *                                  user_arg,
    globus_ftp_client_handle_t *            handle,
    globus_object_t *                       err)
{
    globus_i_replica_management_monitor_t *     monitor;
    static char * myname = "globus_l_replica_management_ftp_list_cb";
    
    monitor = (globus_i_replica_management_monitor_t *) user_arg;
    
    if(err != GLOBUS_NULL)
    {
	monitor->result = globus_error_put(globus_object_copy(err));
    }

    globus_mutex_lock(&monitor->mutex);
    {
	monitor->done = GLOBUS_TRUE;
	globus_cond_signal(&monitor->cond);
    }
    globus_mutex_unlock(&monitor->mutex);
    return;
}


static void
globus_l_replica_management_ftp_cmd_cb(
    void *                                  user_arg,
    globus_ftp_client_handle_t *            handle,
    globus_object_t *                       err)
{
    globus_i_replica_management_monitor_t *     monitor;
    monitor = (globus_i_replica_management_monitor_t *) user_arg;
    
    if(err != GLOBUS_NULL)
    {
	monitor->result = globus_error_put(globus_object_copy(err));
    }

    globus_mutex_lock(&monitor->mutex);
    {
	monitor->done = GLOBUS_TRUE;
	globus_cond_signal(&monitor->cond);
    }
    globus_mutex_unlock(&monitor->mutex);
    return;
}

static void
globus_l_gass_copy_callback(
    void *                                  callback_arg,
    globus_gass_copy_handle_t *             handle,
    globus_object_t *                       err)
{
    globus_i_replica_management_monitor_t *     monitor;
    monitor = (globus_i_replica_management_monitor_t *) callback_arg;
    
    if(err != GLOBUS_NULL)
    {
	monitor->result = globus_error_put(globus_object_copy(err));
    }

    globus_mutex_lock(&monitor->mutex);
    {
	monitor->done = GLOBUS_TRUE;
	globus_cond_signal(&monitor->cond);
    }
    globus_mutex_unlock(&monitor->mutex);
    return;
}

static
globus_result_t
globus_l_replica_management_file_move(
    globus_replica_management_handle_t *              handle,
    char *                                            source_url,
    globus_ftp_client_operationattr_t *               source_attrs,
    char *                                            dest_collection_url,
    globus_replica_catalog_collection_handleattr_t *  dest_collection_attrs,
    char *                                            dest_location_name,
    globus_ftp_client_operationattr_t *               dest_location_attrs,
    char *                                            dest_filename,
    globus_replica_management_overwrite_behavior_t    overwrite_behavior,
    globus_replica_management_preregister_function_t  preregister,
    void *                                            preregister_arg,
    globus_bool_t                                     copy)
{
    globus_result_t                                     rc;
    globus_byte_t                                       buffer[GLOBUS_L_REPLICA_MANAGEMENT_BUFFER_SIZE];
    char *                                              tmp_dest_url = GLOBUS_NULL;
    char *                                              dest_url;
    char *                                              filenames[2];
    char *                                              globus_session_id = GLOBUS_NULL;
    char *                                              rollback_contact;
    globus_i_replica_management_monitor_t               monitor;
    globus_ftp_client_operationattr_t			source_xfer_attrs;
    globus_ftp_client_operationattr_t			dest_xfer_attrs;
    globus_gass_copy_attr_t				source_gass_copy_attrs;
    globus_gass_copy_attr_t				dest_gass_copy_attrs;
    static char * myname = "globus_l_replica_management_file_move";

    rc = globus_i_replica_management_monitor_init(&monitor);

    if(rc != GLOBUS_SUCCESS)
    {
	globus_l_replica_management_release_catalog_handle(handle);
	goto return_result;
    }

    rc = globus_i_replica_catalog_location_get_url(
	handle->replica_handle,
	dest_location_name,
	dest_filename,
	&dest_url);

    if(rc != GLOBUS_SUCCESS)
    {
	globus_l_replica_management_destroy_catalog_handle(handle);
	goto destroy_monitor;
    }

    rc = globus_ftp_client_exists(&handle->ftp_handle,
				dest_url,
				dest_location_attrs,
				globus_l_replica_management_ftp_exists_cb,
				(void *) &monitor);

    if(rc != GLOBUS_SUCCESS)
    {
	globus_l_replica_management_release_catalog_handle(handle);
	goto destroy_url;
    }
    
    globus_mutex_lock(&monitor.mutex);
    {
	while(monitor.done == GLOBUS_FALSE)
	{
	    globus_cond_wait(&monitor.cond, &monitor.mutex);
	}
    }
    globus_mutex_unlock(&monitor.mutex);

    if(monitor.result == GLOBUS_SUCCESS)
    {
	if(overwrite_behavior ==
	   GLOBUS_REPLICA_MANAGEMENT_OVERWRITE_DISALLOW)
	{
	    rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
		"destination file does already exist",
		myname);
	    globus_l_replica_management_release_catalog_handle(handle);
	    goto destroy_url;
	}
	else if(overwrite_behavior ==
		GLOBUS_REPLICA_MANAGEMENT_OVERWRITE_SAFE)
	{
	    globus_session_id = globus_get_unique_session_string();

	    if(globus_session_id == GLOBUS_NULL)
	    {
		rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
		    "Could not allocate internal data structure",
		    myname);
		globus_l_replica_management_release_catalog_handle(handle);
		goto destroy_url;
	    }

	    tmp_dest_url = globus_libc_malloc(sizeof(char) *
					      (strlen(dest_url)
					       + strlen(globus_session_id)
					       + 5));
	    if(tmp_dest_url == GLOBUS_NULL)
	    {
		rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
		    "Could not allocate internal data structure",
		    myname);
		globus_l_replica_management_release_catalog_handle(handle);
		goto destroy_session_id;
	    }

	    strcpy(tmp_dest_url, dest_url);
	    strcat(tmp_dest_url, ".");
	    strcat(tmp_dest_url, globus_session_id);
	}
    }

    rc = globus_i_replica_management_monitor_reset(&monitor);
    
    if(rc != GLOBUS_SUCCESS)
    {
	globus_l_replica_management_release_catalog_handle(handle);
	goto destroy_tmp_url;
    }

    rc = globus_i_replica_management_set_plugin_data(handle,
						     dest_location_name,
						     dest_filename);
    
    if(rc != GLOBUS_SUCCESS)
    {
	globus_l_replica_management_release_catalog_handle(handle);
	goto destroy_tmp_url;
    }

    rc = globus_i_replica_catalog_rollback_info_construct_contact(
	handle->replica_handle,
	dest_location_name,
	dest_filename,
	&rollback_contact);

    if(rc != GLOBUS_SUCCESS)
    {
	globus_l_replica_management_release_catalog_handle(handle);
	goto destroy_tmp_url;
    }

    switch(overwrite_behavior)
    {
    case GLOBUS_REPLICA_MANAGEMENT_OVERWRITE_SAFE:
	rc = globus_i_replica_catalog_rollback_info_create(
	    handle->replica_handle,
	    dest_location_name,
	    dest_filename,
	    copy == GLOBUS_FALSE ? GLOBUS_I_REPLICA_MANAGEMENT_PUBLISH :
	    GLOBUS_I_REPLICA_MANAGEMENT_COPY,
	    GLOBUS_I_REPLICA_MANAGEMENT_SAFE_START,
	    globus_session_id,
	    source_url);
	break;
    case GLOBUS_REPLICA_MANAGEMENT_OVERWRITE_DISALLOW:
    case GLOBUS_REPLICA_MANAGEMENT_OVERWRITE_UNSAFE:
	rc = globus_i_replica_catalog_rollback_info_create(
	    handle->replica_handle,
	    dest_location_name,
	    dest_filename,
	    copy == GLOBUS_FALSE ? GLOBUS_I_REPLICA_MANAGEMENT_PUBLISH :
	    GLOBUS_I_REPLICA_MANAGEMENT_COPY,
	    GLOBUS_I_REPLICA_MANAGEMENT_UNSAFE_START,
	    globus_session_id,
	    source_url);
	break;
    }
    
    if(rc != GLOBUS_SUCCESS)
    {
	globus_l_replica_management_release_catalog_handle(handle);
	goto destroy_contact;
    }

    rc = globus_ftp_client_operationattr_init(&source_xfer_attrs);
    if (rc != GLOBUS_SUCCESS)
    {
	globus_l_replica_management_destroy_catalog_handle(handle);
	goto destroy_contact;
    }

    rc = globus_ftp_client_operationattr_copy( &source_xfer_attrs,
						source_attrs);
    if (rc != GLOBUS_SUCCESS)
    {
	globus_l_replica_management_release_catalog_handle(handle);
	goto destroy_source_xfer_attrs;
    }

    rc = globus_ftp_client_operationattr_init(&dest_xfer_attrs);
    if (rc != GLOBUS_SUCCESS)
    {
	globus_l_replica_management_destroy_catalog_handle(handle);
	goto destroy_source_xfer_attrs;
    }

    rc = globus_ftp_client_operationattr_copy( &dest_xfer_attrs,
						dest_location_attrs);
    if (rc != GLOBUS_SUCCESS)
    {
	globus_l_replica_management_release_catalog_handle(handle);
	goto destroy_dest_xfer_attrs;
    }

    rc = globus_gass_copy_attr_init(&source_gass_copy_attrs);
    if (rc != GLOBUS_SUCCESS)
    {
	globus_l_replica_management_release_catalog_handle(handle);
	goto destroy_dest_xfer_attrs;
    }

    rc = globus_gass_copy_attr_set_ftp(&source_gass_copy_attrs,
				    &source_xfer_attrs);
    if (rc != GLOBUS_SUCCESS)
    {
	globus_l_replica_management_release_catalog_handle(handle);
	goto destroy_dest_xfer_attrs;
    }

    rc = globus_gass_copy_attr_init(&dest_gass_copy_attrs);
    if (rc != GLOBUS_SUCCESS)
    {
	globus_l_replica_management_release_catalog_handle(handle);
	goto destroy_dest_xfer_attrs;
    }

    rc = globus_gass_copy_attr_set_ftp(&dest_gass_copy_attrs,
				    &dest_xfer_attrs);
    if (rc != GLOBUS_SUCCESS)
    {
	globus_l_replica_management_release_catalog_handle(handle);
	goto destroy_dest_xfer_attrs;
    }

    rc = globus_gass_copy_register_url_to_url(
	    &handle->gass_copy_handle,
	    source_url,
	    &source_gass_copy_attrs,
	    tmp_dest_url ? tmp_dest_url : dest_url, 
	    &dest_gass_copy_attrs,
	    globus_l_gass_copy_callback,
	    (void *) &monitor);
    if (rc != GLOBUS_SUCCESS)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ROLLBACK_ERROR(
	    rollback_contact,
	    globus_error_get(rc));
	globus_l_replica_management_destroy_catalog_handle(handle);
	goto destroy_dest_xfer_attrs;
    }

    globus_mutex_lock(&monitor.mutex);
    {
	while(monitor.done == GLOBUS_FALSE)
	{
	    globus_cond_wait(&monitor.cond, &monitor.mutex);
	}
    }
    globus_mutex_unlock(&monitor.mutex);

    if(monitor.result != GLOBUS_SUCCESS)
    {
	rc = monitor.result;
	rc = GLOBUS_REPLICA_MANAGEMENT_ROLLBACK_ERROR(
	    rollback_contact,
	    globus_error_get(rc));
	globus_l_replica_management_release_catalog_handle(handle);
	goto destroy_dest_xfer_attrs;
    }

    if(tmp_dest_url != GLOBUS_NULL)
    {
	rc = GLOBUS_I_REPLICA_CATALOG_ROLLBACK_INFO_UPDATE(
	    handle->replica_handle,
	    dest_location_name,
	    dest_filename,
	    GLOBUS_I_REPLICA_MANAGEMENT_NEED_FILE_MOVE,
	    GLOBUS_NULL);
	
	if(rc != GLOBUS_SUCCESS)
	{
	    rc = GLOBUS_REPLICA_MANAGEMENT_ROLLBACK_ERROR(
		rollback_contact,
		globus_error_get(rc));
	    globus_l_replica_management_destroy_catalog_handle(handle);
	    goto destroy_dest_xfer_attrs;
	}
	
	globus_i_replica_management_monitor_reset(&monitor);
	
	rc = globus_ftp_client_move(
	    &handle->ftp_handle,
	    tmp_dest_url,
	    dest_url,
	    dest_location_attrs,
	    globus_l_replica_management_ftp_cmd_cb,
	    (void *) &monitor);
	
	if(rc != GLOBUS_SUCCESS)
	{
	    rc = GLOBUS_REPLICA_MANAGEMENT_ROLLBACK_ERROR(
		rollback_contact,
		globus_error_get(rc));
	    globus_l_replica_management_release_catalog_handle(handle);
	    goto destroy_dest_xfer_attrs;
	}
	
	globus_mutex_lock(&monitor.mutex);
	{
	    while(monitor.done == GLOBUS_FALSE)
	    {
		globus_cond_wait(&monitor.cond, &monitor.mutex);
	    }
	}
	globus_mutex_unlock(&monitor.mutex);
	
	if(monitor.result != GLOBUS_SUCCESS)
	{
	    rc = monitor.result;
	    rc = GLOBUS_REPLICA_MANAGEMENT_ROLLBACK_ERROR(
		rollback_contact,
		globus_error_get(rc));
	    globus_l_replica_management_release_catalog_handle(handle);
	    goto destroy_dest_xfer_attrs;
	}
    }

    rc = GLOBUS_I_REPLICA_CATALOG_ROLLBACK_INFO_UPDATE(
	handle->replica_handle,
	dest_location_name,
	dest_filename,
	GLOBUS_I_REPLICA_MANAGEMENT_NEED_PREREGISTER,
	GLOBUS_NULL);

    if(rc != GLOBUS_SUCCESS)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ROLLBACK_ERROR(
	    rollback_contact,
	    globus_error_get(rc));
	globus_l_replica_management_destroy_catalog_handle(handle);
	goto destroy_dest_xfer_attrs;
    }

    
    if(preregister != GLOBUS_NULL)
    {
	rc = preregister(preregister_arg,
			 dest_url);
	if(rc != GLOBUS_SUCCESS)
	{
	    globus_l_replica_management_release_catalog_handle(handle);
	    goto destroy_dest_xfer_attrs;
	}
    }
    
    filenames[0] = dest_filename;
    filenames[1] = GLOBUS_NULL;
	
    if(copy == GLOBUS_FALSE)
    {
	rc = GLOBUS_I_REPLICA_CATALOG_ROLLBACK_INFO_UPDATE(
	    handle->replica_handle,
	    dest_location_name,
	    dest_filename,
	    GLOBUS_I_REPLICA_MANAGEMENT_NEED_COLLECTION_ADD,
	    GLOBUS_NULL);
	
	if(rc != GLOBUS_SUCCESS)
	{
	    rc = GLOBUS_REPLICA_MANAGEMENT_ROLLBACK_ERROR(
		rollback_contact,
		globus_error_get(rc));
	    globus_l_replica_management_destroy_catalog_handle(handle);
	    goto destroy_dest_xfer_attrs;
	}

	rc = GLOBUS_REPLICA_CATALOG_COLLECTION_ADD_FILENAMES(
	    handle->replica_handle,
	    filenames,
	    GLOBUS_TRUE);
	
	if(rc != GLOBUS_SUCCESS)
	{
	    rc = GLOBUS_REPLICA_MANAGEMENT_ROLLBACK_ERROR(
		rollback_contact,
		globus_error_get(rc));
	    globus_l_replica_management_destroy_catalog_handle(handle);
	    goto destroy_dest_xfer_attrs;
	}
    }

    rc = GLOBUS_I_REPLICA_CATALOG_ROLLBACK_INFO_UPDATE(
	handle->replica_handle,
	dest_location_name,
	dest_filename,
	GLOBUS_I_REPLICA_MANAGEMENT_NEED_LOCATION_ADD,
	GLOBUS_NULL);

    if(rc != GLOBUS_SUCCESS)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ROLLBACK_ERROR(
	    rollback_contact,
	    globus_error_get(rc));
	globus_l_replica_management_destroy_catalog_handle(handle);
	goto destroy_dest_xfer_attrs;
    }

    rc = GLOBUS_REPLICA_CATALOG_LOCATION_ADD_FILENAMES(
	handle->replica_handle,
	dest_location_name,
	filenames,
	GLOBUS_TRUE);

    if(rc != GLOBUS_SUCCESS)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ROLLBACK_ERROR(
	    rollback_contact,
	    globus_error_get(rc));
	globus_l_replica_management_destroy_catalog_handle(handle);
	goto destroy_dest_xfer_attrs;
    }

    rc = GLOBUS_I_REPLICA_CATALOG_ROLLBACK_INFO_DELETE(
	handle->replica_handle,
	dest_location_name,
	dest_filename);

    if(rc != GLOBUS_SUCCESS)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ROLLBACK_ERROR(
	    rollback_contact,
	    globus_error_get(rc));
	
	globus_l_replica_management_destroy_catalog_handle(handle);
	goto destroy_dest_xfer_attrs;
    }
    
    rc = globus_l_replica_management_release_catalog_handle(handle);

 destroy_dest_xfer_attrs:
    globus_ftp_client_operationattr_destroy(&dest_xfer_attrs);
 destroy_source_xfer_attrs:
    globus_ftp_client_operationattr_destroy(&source_xfer_attrs);
 destroy_contact:
    globus_libc_free(rollback_contact);
 destroy_tmp_url:
    if(tmp_dest_url != GLOBUS_NULL)
    {
	globus_libc_free(tmp_dest_url);
    }
 destroy_session_id:
    if(globus_session_id != GLOBUS_NULL)
    {
	globus_libc_free(globus_session_id);
    } 
 destroy_url:
    globus_libc_free(dest_url);
 destroy_monitor:
    globus_i_replica_management_monitor_destroy(&monitor);
 return_result:
    return rc;
}


static globus_l_replica_management_restart_state_t
globus_l_replica_management_get_state(
    char *                              state)
{
    if(!strcmp(state, GLOBUS_I_REPLICA_MANAGEMENT_PHYSICAL_DELETE_START))
    {
	return GLOBUS_L_REPLICA_MANAGEMENT_PHYSICAL_DELETE_START;
    }
    else if(!strcmp(state, GLOBUS_I_REPLICA_MANAGEMENT_LOGICAL_DELETE_START))
    {
	return GLOBUS_L_REPLICA_MANAGEMENT_LOGICAL_DELETE_START;
    }
    else if(!strcmp(state, GLOBUS_I_REPLICA_MANAGEMENT_NEED_PHYSICAL_DELETE))
    {
	return GLOBUS_L_REPLICA_MANAGEMENT_NEED_PHYSICAL_DELETE;
    }
    else if(!strcmp(state, GLOBUS_I_REPLICA_MANAGEMENT_SAFE_START))
    {
	return GLOBUS_L_REPLICA_MANAGEMENT_SAFE_START;
    }
    else if(!strcmp(state, GLOBUS_I_REPLICA_MANAGEMENT_UNSAFE_START))
    {
	return GLOBUS_L_REPLICA_MANAGEMENT_UNSAFE_START;
    }
    else if(!strcmp(state, GLOBUS_I_REPLICA_MANAGEMENT_NEED_FILE_MOVE))
    {
	return GLOBUS_L_REPLICA_MANAGEMENT_NEED_FILE_MOVE;
    }
    else if(!strcmp(state, GLOBUS_I_REPLICA_MANAGEMENT_NEED_PREREGISTER))
    {
	return GLOBUS_L_REPLICA_MANAGEMENT_NEED_PREREGISTER;
    }
    else if(!strcmp(state, GLOBUS_I_REPLICA_MANAGEMENT_NEED_COLLECTION_ADD))
    {
	return GLOBUS_L_REPLICA_MANAGEMENT_NEED_COLLECTION_ADD;
    }
    else if(!strcmp(state, GLOBUS_I_REPLICA_MANAGEMENT_NEED_LOCATION_ADD))
    {
	return GLOBUS_L_REPLICA_MANAGEMENT_NEED_LOCATION_ADD;
    }

    return -1;
}


static globus_result_t
globus_l_replica_management_create_dir(
    globus_replica_management_handle_t *    handle,
    globus_ftp_client_operationattr_t *     ftp_attrs,
    char *				    root_dir_url)
{
    globus_result_t                         rc;
    globus_url_t                            url;
    globus_i_replica_management_monitor_t   monitor;
    globus_byte_t                           buffer[GLOBUS_L_REPLICA_MANAGEMENT_BUFFER_SIZE];
    char *                                  tmp_url;
    char *                                  pos;
    char                                    replaced_char;
    static char * myname = "globus_l_replica_management_create_dir";
    
    if(handle == GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "NULL handle",
	    myname);
	goto return_result;
    }

    if(ftp_attrs == GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "NULL FTP attributes",
	    myname);
	goto return_result;
    }

    if(root_dir_url == GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "NULL URL",
	    myname);
	goto return_result;
    }
    
    if(globus_url_parse(root_dir_url,&url))
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "Failed to parse URL",
	    myname);
	goto return_result;
    }


    if(url.url_path == GLOBUS_NULL)
    {
	rc = GLOBUS_SUCCESS;
	goto destroy_url;
    }
    
    tmp_url = strdup(root_dir_url);

    if(tmp_url == GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "Could not allocate internal data structure",
	    myname);
	goto destroy_url;
    }

    rc = globus_i_replica_management_monitor_init(&monitor);

    if(rc != GLOBUS_SUCCESS)
    {
	goto destroy_tmp_url;
    }

    
    pos = strstr(tmp_url, url.url_path);

    while(*pos != '\0' && *pos == '/')
    {
	pos++;
    }

    while(*pos != '\0')
    {
	if(*pos == '/' || *(pos + 1) == '\0')
	{
	    replaced_char = *(pos + 1);
	    *(pos + 1) = '\0';

            rc = globus_ftp_client_exists(
                &handle->ftp_handle,
                tmp_url,
                ftp_attrs,
                globus_l_replica_management_ftp_exists_cb,
                (void *) &monitor);
	    
	    if(rc != GLOBUS_SUCCESS)
	    {
		goto destroy_monitor;
	    }

	    globus_mutex_lock(&monitor.mutex);
	    {
		while(monitor.done == GLOBUS_FALSE)
		{
		    globus_cond_wait(&monitor.cond, &monitor.mutex);
		}
	    }
	    globus_mutex_unlock(&monitor.mutex);

	    /* if directory/file didn't exist, try to create it
	     * would be safer with real errors...  
	     */
	    
	    if(monitor.result != GLOBUS_SUCCESS)
	    {
		rc = globus_i_replica_management_monitor_reset(&monitor);

		if(rc != GLOBUS_SUCCESS)
		{
		    goto destroy_monitor;
		}
		
		rc = globus_ftp_client_mkdir(
		    &handle->ftp_handle,
		    tmp_url,
		    ftp_attrs,
		    globus_l_replica_management_ftp_cmd_cb,
		    (void *) &monitor);

		if(rc != GLOBUS_SUCCESS)
		{
		    goto destroy_monitor;
		}
		
		
		globus_mutex_lock(&monitor.mutex);
		{
		    while(monitor.done == GLOBUS_FALSE)
		    {
			globus_cond_wait(&monitor.cond, &monitor.mutex);
		    }
		}
		globus_mutex_unlock(&monitor.mutex);

		if(monitor.result != GLOBUS_SUCCESS)
		{
		    rc = monitor.result;
		    goto destroy_monitor;
		}
	    }

	    *(pos + 1) = replaced_char;

	    rc = globus_i_replica_management_monitor_reset(&monitor);
	    
	    if(rc != GLOBUS_SUCCESS)
	    {
		goto destroy_monitor;
	    }
	}
	pos++;
    }

    rc = GLOBUS_SUCCESS;

 destroy_monitor:
    globus_i_replica_management_monitor_destroy(&monitor);
 destroy_tmp_url:
    globus_libc_free(tmp_url);
 destroy_url:
    globus_url_destroy(&url);
 return_result:
    return rc;
}


globus_result_t
globus_replica_management_get_rollback_contact(
    globus_result_t                     result,
    char **                             contact_string)
{
    globus_result_t                     rc;
    globus_object_t *                   err;
    static char * myname = "globus_replica_management_get_rollback_contact";
    
    if(result == GLOBUS_SUCCESS)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
		"No rollback contact found",
		myname);
	goto return_result;
    }

    err = globus_error_get(result);

    if(err == GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "Unable to obtain error object",
	    myname);
	goto return_result;
    }

    if(globus_object_get_type(err) != GLOBUS_ERROR_TYPE_ROLLBACK_CONTACT)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "No rollback contact found",
	    myname);
	goto put_error;
    }

    *contact_string = globus_object_printable_to_string(err);

    if(*contact_string == GLOBUS_NULL)
    {
	rc = GLOBUS_REPLICA_MANAGEMENT_ERROR(
	    "Could not allocate internal data structure",
	    myname);
	goto put_error;
    }

    rc = GLOBUS_SUCCESS;

 put_error:
    globus_error_put(err);
 return_result:
    return rc;
    
}




